from .._guidance import guidance
from ._block import block
from ._set_attribute import set_attribute

nodisp_start = "<||_#NODISP_||>"
nodisp_end = "<||_/NODISP_||>"
span_start = "<||_html:<span style='background-color: rgba(255, 180, 0, 0.3); border-radius: 3px;'>_||>"
span_end = "<||_html:</span>_||>"

@guidance
def role_opener(lm, role_name, **kwargs):
    indent = getattr(lm, "indent_roles", True)


    # Block start container (centers elements)
    if indent:
        lm += f"<||_html:<div style='display: flex; border-bottom: 1px solid rgba(127, 127, 127, 0.2);  justify-content: center; align-items: center;'><div style='flex: 0 0 80px; opacity: 0.5;'>{role_name.lower()}</div><div style='flex-grow: 1; padding: 5px; padding-top: 10px; padding-bottom: 10px; margin-top: 0px; white-space: pre-wrap; margin-bottom: 0px;'>_||>"

    # Start of either debug or HTML no disp block
    if indent:
        lm += nodisp_start
    else:
        lm += span_start
    
    # TODO [HN]: Temporary change while I instrument chat_template in transformers only.
    # Eventually have all models use chat_template.
    if hasattr(lm, "get_role_start"):
        lm += lm.get_role_start(role_name, **kwargs)
    elif hasattr(lm, "chat_template"):
        lm += lm.chat_template.get_role_start(role_name)
    else:
        raise Exception(
            f"You need to use a chat model in order the use role blocks like `with {role_name}():`! Perhaps you meant to use the {type(lm).__name__}Chat class?"
        )

    # End of either debug or HTML no disp block
    if indent:
        lm += nodisp_end
    else:
        lm += span_end

    return lm


@guidance
def role_closer(lm, role_name, **kwargs):
    indent = getattr(lm, "indent_roles", True)
    # Start of either debug or HTML no disp block
    if indent:
        lm += nodisp_start
    else:
        lm += span_start

    # TODO [HN]: Temporary change while I instrument chat_template in transformers only.
    # Eventually have all models use chat_template.
    if hasattr(lm, "get_role_end"):
        lm += lm.get_role_end(role_name)
    elif hasattr(lm, "chat_template"):
        lm += lm.chat_template.get_role_end(role_name)

    # End of either debug or HTML no disp block
    if indent:
        lm += nodisp_end
    else:
        lm += span_end

    # End of top container
    if indent:
        lm += "<||_html:</div></div>_||>"

    return lm

# TODO HN: Add a docstring to better describe arbitrary role functions
def role(role_name, text=None, **kwargs):
    if text is None:
        return block(
            opener=role_opener(role_name, **kwargs),
            closer=role_closer(role_name, **kwargs),
        )
    else:
        assert False
        # return self.append(open_text + text + close_text)


def system(text=None, **kwargs):
    """Indicate the 'system' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of a 'system' block, which
    provides background information to the LLM.

        >>> with system():
        >>>     lm += "A system prompt"

    """
    return role("system", text, **kwargs)


def user(text=None, **kwargs):
    """Indicate the 'user' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of a 'user' block, which
    provides input to the LLM from the user.

        >>> with user():
        >>>     lm += "What the user said"

    """
    return role("user", text, **kwargs)


def assistant(text=None, **kwargs):
    """Indicate the 'assistant' prompt

    A convention has grown up around 'chat' APIs that
    prompts are split into three parts: system, user
    and assistant.
    This indicates the start of an 'assistant' block, which
    marks LLM response (or where the LLM will generate
    the next response).

        >>> with assistant():
        >>>     lm += gen(name="model_output", max_tokens=20)

    """
    return role("assistant", text, **kwargs)


def function(text=None, **kwargs):
    return role("function", text, **kwargs)


def instruction(text=None, **kwargs):
    return role("instruction", text, **kwargs)


def indent_roles(indent=True):
    return set_attribute("indent_roles", indent)



"""Shared utility functions for module."""

import os
from pathlib import Path

def lib_bench_dir() -> Path:
    """Library directory to store configurations and cached assets for benchmarking.

    If the library directory does not exist, it is created as a side effect.

    The library bench directory path can also be set via env var `GUIDANCE_BENCH_DIR`.

    Returns:
        Path: Library's directory path for benchmarking.
    """

    env_lib_path = os.environ.get("GUIDANCE_BENCH_DIR", None)
    if env_lib_path is None:
        lib_path = Path.home() / ".guidance-bench"
    else:
        lib_path = Path(env_lib_path)
    Path.mkdir(lib_path, parents=True, exist_ok=True)

    return lib_path



import re
from ._model import Chat, Instruct
from ._grammarless import Grammarless, GrammarlessEngine
import tiktoken
import os

_image_token_pattern = re.compile(r"<\|_image:(.*)\|>")


class GoogleAIEngine(GrammarlessEngine):
    def __init__(
        self,
        model,
        tokenizer,
        api_key,
        max_streaming_tokens,
        timeout,
        compute_log_probs,
        **kwargs,
    ):
        try:
            import google.generativeai as genai
        except ModuleNotFoundError:
            raise Exception(
                "Please install the Google AI Studio(makersuite.google.com) package using `pip install google-generativeai google-ai-generativelanguage` in order to use guidance.models.GoogleAI!"
            )

        assert (
            not compute_log_probs
        ), "We don't support compute_log_probs=True yet for GoogleAIEngine!"

        if api_key is None:
            api_key = os.environ.get("GOOGLEAI_API_KEY")

        genai.configure(api_key=api_key)

        # Gemini does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")
        self.model_name = model

        self.model_obj = genai.GenerativeModel(self.model_name, **kwargs)

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)


class GoogleAI(Grammarless):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        max_streaming_tokens=None,
        timeout=0.5,
        compute_log_probs=False,
        **kwargs,
    ):
        """Build a new GoogleAI model object that represents a model in a given state."""

        # if we are called directly (as opposed to through super()) then we convert ourselves to a more specific subclass if possible
        if self.__class__ is GoogleAI:
            found_subclass = None

            # chat
            found_subclass = GoogleAIChat  # we assume all models are chat right now

            # instruct
            # elif "instruct" in model:
            #     found_subclass = GoogleAIInstruct

            # # regular completion
            # else:
            #     found_subclass = GoogleAICompletion

            # convert to any found subclass
            self.__class__ = found_subclass
            found_subclass.__init__(
                self,
                model,
                tokenizer=None,
                echo=True,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=False,
                **kwargs,
            )
            return  # we return since we just ran init above and don't need to run again

        # this allows us to use a single constructor for all our subclasses
        engine_map = {GoogleAIChat: GoogleAIChatEngine}

        super().__init__(
            engine=engine_map[self.__class__](
                model=model,
                tokenizer=tokenizer,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                **kwargs,
            ),
            echo=echo,
        )


class GoogleAIChatEngine(GoogleAIEngine):
    def _generator(self, prompt, temperature):

        # find the system text
        pos = 0
        system_start = b"<|im_start|>system\n"
        user_start = b"<|im_start|>user\n"
        assistant_start = b"<|im_start|>assistant\n"
        role_end = b"<|im_end|>"
        # system_start_pos = prompt.startswith(system_start)

        # find the system text
        system_text = b""
        if prompt.startswith(system_start):
            pos += len(system_start)
            system_end_pos = prompt.find(role_end)
            system_text = prompt[pos:system_end_pos]
            pos = system_end_pos + len(role_end)

        # find the user/assistant pairs
        messages = []
        valid_end = False
        while True:

            # find the user text
            if prompt[pos:].startswith(user_start):
                pos += len(user_start)
                end_pos = prompt[pos:].find(role_end)
                if end_pos < 0:
                    break
                messages.append(
                    dict(
                        role="user",
                        content=prompt[pos : pos + end_pos].decode("utf8"),
                    )
                )
                pos += end_pos + len(role_end)
            elif prompt[pos:].startswith(assistant_start):
                pos += len(assistant_start)
                end_pos = prompt[pos:].find(role_end)
                if end_pos < 0:
                    valid_end = True
                    break
                messages.append(
                    dict(
                        role="model",
                        content=prompt[pos : pos + end_pos].decode("utf8"),
                    )
                )
                pos += end_pos + len(role_end)
            else:
                raise Exception(
                    "It looks like your prompt is not a well formed chat prompt! Please enclose all model state appends inside chat role blocks like `user()` or `assistant()`."
                )

        self._data = prompt[:pos]

        assert len(messages) > 0, "Bad chat format! No chat blocks were defined."
        assert (
            messages[-1]["role"] == "user"
        ), "Bad chat format! There must be a user() role before the last assistant() role."
        assert valid_end, "Bad chat format! You must generate inside assistant() roles."

        # TODO: don't make a new session on every call
        # last_user_text = messages.pop().content

        return self._start_generator(system_text.decode("utf8"), messages, temperature)

        # kwargs = {}
        # if self.max_streaming_tokens is not None:
        #     kwargs["max_output_tokens"] = self.max_streaming_tokens
        # generator = chat_session.send_message_streaming(last_user_text, temperature=temperature, **kwargs)

        # for chunk in generator:
        #     yield chunk.text.encode("utf8")

    def _start_chat(self, system_text, messages):
        assert (
            system_text == ""
        ), "We don't support passing system text to Gemini models (yet?)!"
        out = self.model_obj.start_chat(history=messages)
        return out

    def _start_generator(self, system_text, messages, temperature):
        from google.ai.generativelanguage import Content, Part, Blob

        # last_user_text = messages[-1]["content"]
        formated_messages = []
        for m in messages:
            raw_parts = _image_token_pattern.split(m["content"])
            parts = []
            for i in range(0, len(raw_parts), 2):

                # append the text portion
                if len(raw_parts[i]) > 0:
                    parts.append(Part(text=raw_parts[i]))

                # append any image
                if i + 1 < len(raw_parts):
                    # parts.append(Part.from_image(Image.from_bytes(self[raw_parts[i+1]])))
                    parts.append(
                        Part(
                            inline_data=Blob(
                                mime_type="image/jpeg", data=self[raw_parts[i + 1]]
                            )
                        )
                    )
            formated_messages.append(Content(role=m["role"], parts=parts))
        last_user_parts = (
            formated_messages.pop()
        )  # remove the last user stuff that goes in send_message (and not history)

        chat_session = self.model_obj.start_chat(
            history=formated_messages,
        )

        generation_config = {"temperature": temperature}
        if self.max_streaming_tokens is not None:
            generation_config["max_output_tokens"] = self.max_streaming_tokens
        generator = chat_session.send_message(
            last_user_parts, generation_config=generation_config, stream=True
        )

        for chunk in generator:
            yield chunk.candidates[0].content.parts[0].text.encode("utf8")


class GoogleAIChat(GoogleAI, Chat):
    pass



# import functions that can be called directly
from ._gen import gen, call_tool, will_gen
from ._image import image
from ._capture import capture

# core grammar functions
from .._grammar import select
from .._grammar import commit_point
from .._grammar import with_temperature
from .._grammar import string
from .._grammar import token_limit
from .._grammar import byte_range

# context blocks
from ._block import block
from ._role import role, system, assistant, user, function, instruction, indent_roles
from ._format import monospace
from ._silent import silent
from ._set_var import set_var
from ._set_attribute import set_attribute

# from ..models._model import context_free

# stateless library functions
from ._any_char import any_char
from ._sequences import one_or_more, zero_or_more, at_most_n_repeats, exactly_n_repeats, sequence
from ._char_range import char_range
from ._char_set import char_set
from ._prefix_tree import prefix_tree
from ._substring import substring
from ._regex import regex
from ._optional import optional
from ._tool import Tool
from ._any_char_but import any_char_but

from ._json import json



from ._lite_llm import LiteLLMEngine, LiteLLM, LiteLLMCompletion, LiteLLMInstruct


class Cohere(LiteLLM):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        timeout=0.5,
        compute_log_probs=False,
        max_streaming_tokens=1000,
    ):
        """Build a new Anthropic model object that represents a model in a given state."""
        try:
            import tokenizers
        except ModuleNotFoundError:
            raise Exception(
                "Please install the HuggingFace tokenizers package using `pip install tokenizers -U` in order to use guidance.models.Cohere!"
            )

        # get the tokenizer
        if tokenizer is None:
            try:
                tokenizer = tokenizers.Tokenizer.from_pretrained("Cohere/" + model)
            except:
                tokenizer = tokenizers.Tokenizer.from_pretrained(
                    "Cohere/command-nightly"
                )

        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            timeout=timeout,
            max_streaming_tokens=max_streaming_tokens,
            compute_log_probs=compute_log_probs,
        )


class CohereCompletion(Cohere, LiteLLMCompletion):
    pass


class CohereInstruct(Cohere, LiteLLMInstruct):
    pass



import inspect
import warnings

from typing import Dict, Union


class ChatTemplate:
    """Contains template for all chat and instruct tuned models."""

    def get_role_start(self, role_name: str, **kwargs):
        raise NotImplementedError(
            "You need to use a ChatTemplate subclass that overrides the get_role_start method"
        )

    def get_role_end(self, role_name: Union[str, None] = None):
        raise NotImplementedError(
            "You need to use a ChatTemplate subclass that overrides the get_role_start method"
        )


class ChatTemplateCache:
    def __init__(self):
        self._cache: Dict[str, ChatTemplate] = {}

    def __getitem__(self, key: str) -> ChatTemplate:
        key_compact = key.replace(" ", "")
        return self._cache[key_compact]

    def __setitem__(self, key: str, value):
        key_compact = key.replace(" ", "")
        self._cache[key_compact] = value

    def __contains__(self, key: str):
        key_compact = key.replace(" ", "")
        return key_compact in self._cache


# Feels weird having to instantiate this, but it's a singleton for all purposes
# TODO [HN]: Add an alias system so we can instantiate with other simple keys (e.g. "llama2" instead of the full template string)
CHAT_TEMPLATE_CACHE = ChatTemplateCache()


class UnsupportedRoleException(Exception):
    def __init__(self, role_name, instance):
        self.role_name = role_name
        self.instance = instance
        super().__init__(self._format_message())

    def _format_message(self):
        return f"Role {self.role_name} is not supported by the {self.instance.__class__.__name__} chat template. "


def load_template_class(chat_template=None):
    """Utility method to find the best chat template.

    Order of precedence:
    - If it's a chat template class, use it directly
    - If it's a string, check the cache of popular model templates
    - If it's a string and not in the cache, try to create a class dynamically
    - [TODO] If it's a string and can't be created, default to ChatML and raise a warning
    - If it's None, default to ChatML and raise a warning
    """
    if inspect.isclass(chat_template) and issubclass(chat_template, ChatTemplate):
        if chat_template is ChatTemplate:
            raise Exception(
                "You can't use the base ChatTemplate class directly. Create or use a subclass instead."
            )
        return chat_template

    elif isinstance(chat_template, str):
        # First check the cache of popular model types
        # TODO: Expand keys of cache to include aliases for popular model types (e.g. "llama2, phi3")
        # Can possibly accomplish this with an "aliases" dictionary that maps all aliases to the canonical key in cache
        if chat_template in CHAT_TEMPLATE_CACHE:
            return CHAT_TEMPLATE_CACHE[chat_template]
        # TODO: Add logic here to try to auto-create class dynamically via _template_class_from_string method

    # Only warn when a user provided a chat template that we couldn't load
    if chat_template is not None:
        warnings.warn(
            f"""Chat template {chat_template} was unable to be loaded directly into guidance.
                        Defaulting to the ChatML format which may not be optimal for the selected model. 
                        For best results, create and pass in a `guidance.ChatTemplate` subclass for your model."""
        )

    # By default, use the ChatML Template. Warnings to user will happen downstream only if they use chat roles.
    return ChatMLTemplate


def _template_class_from_string(template_str):
    """Utility method to try to create a chat template class from a string."""
    # TODO: Try to build this, perhaps based on passing unit tests we create?
    pass


# CACHE IMPLEMENTATIONS:

# --------------------------------------------------
# @@@@ ChatML @@@@
# --------------------------------------------------
# Note that all grammarless models will default to this syntax, since we typically send chat formatted messages.
chatml_template = "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}{% endfor %}"


class ChatMLTemplate(ChatTemplate):
    template_str = chatml_template

    def get_role_start(self, role_name):
        return f"<|im_start|>{role_name}\n"

    def get_role_end(self, role_name=None):
        return "<|im_end|>\n"


CHAT_TEMPLATE_CACHE[chatml_template] = ChatMLTemplate


# --------------------------------------------------
# @@@@ Llama-2 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/meta-llama/Llama-2-7b-chat-hf/blob/main/tokenizer_config.json#L12
llama2_template = "{% if messages[0]['role'] == 'system' %}{% set loop_messages = messages[1:] %}{% set system_message = messages[0]['content'] %}{% else %}{% set loop_messages = messages %}{% set system_message = false %}{% endif %}{% for message in loop_messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if loop.index0 == 0 and system_message != false %}{% set content = '<<SYS>>\\n' + system_message + '\\n<</SYS>>\\n\\n' + message['content'] %}{% else %}{% set content = message['content'] %}{% endif %}{% if message['role'] == 'user' %}{{ bos_token + '[INST] ' + content.strip() + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ ' '  + content.strip() + ' ' + eos_token }}{% endif %}{% endfor %}"


class Llama2ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = llama2_template

    def get_role_start(self, role_name):
        if role_name == "system":
            return "[INST] <<SYS>>\n"
        elif role_name == "user":
            return "<s>[INST] "
        elif role_name == "assistant":
            return " "
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "system":
            return "\n<</SYS>>"
        elif role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return " </s>"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[llama2_template] = Llama2ChatTemplate


# --------------------------------------------------
# @@@@ Llama-3 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/meta-llama/Meta-Llama-3-8B-Instruct/blob/main/tokenizer_config.json#L2053
llama3_template = "{% set loop_messages = messages %}{% for message in loop_messages %}{% set content = '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n'+ message['content'] | trim + '<|eot_id|>' %}{% if loop.index0 == 0 %}{% set content = bos_token + content %}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' }}{% endif %}"


class Llama3ChatTemplate(ChatTemplate):
    # available_roles = ["system", "user", "assistant"]
    template_str = llama3_template

    def get_role_start(self, role_name):
        if role_name == "system":
            return "<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\n"
        elif role_name == "user":
            return "<|start_header_id|>user<|end_header_id|>\n\n"
        elif role_name == "assistant":
            return "<|start_header_id|>assistant<|end_header_id|>\n\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        return "<|eot_id|>"


CHAT_TEMPLATE_CACHE[llama3_template] = Llama3ChatTemplate

# --------------------------------------------------
# @@@@ Phi-3 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/microsoft/Phi-3-mini-4k-instruct/blob/main/tokenizer_config.json#L119
phi3_mini_template = "{% for message in messages %}{% if message['role'] == 'system' %}{{'<|system|>\n' + message['content'] + '<|end|>\n'}}{% elif message['role'] == 'user' %}{{'<|user|>\n' + message['content'] + '<|end|>\n'}}{% elif message['role'] == 'assistant' %}{{'<|assistant|>\n' + message['content'] + '<|end|>\n'}}{% endif %}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% else %}{{ eos_token }}{% endif %}"
phi3_mini_llamacpp_template = "{{ bos_token }}{% for message in messages %}{% if (message['role'] == 'user') %}{{'<|user|>' + '\n' + message['content'] + '<|end|>' + '\n' + '<|assistant|>' + '\n'}}{% elif (message['role'] == 'assistant') %}{{message['content'] + '<|end|>' + '\n'}}{% endif %}{% endfor %}"


class Phi3MiniChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = phi3_mini_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<|user|>\n"
        elif role_name == "assistant":
            return "<|assistant|>\n"
        elif role_name == "system":
            return "<|system|>\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        return "<|end|>\n"


CHAT_TEMPLATE_CACHE[phi3_mini_template] = Phi3MiniChatTemplate
CHAT_TEMPLATE_CACHE[phi3_mini_llamacpp_template] = Phi3MiniChatTemplate

# https://huggingface.co/microsoft/Phi-3-small-8k-instruct/blob/main/tokenizer_config.json
phi3_small_template = "{{ bos_token }}{% for message in messages %}{{'<|' + message['role'] + '|>' + '\n' + message['content'] + '<|end|>\n' }}{% endfor %}{% if add_generation_prompt %}{{ '<|assistant|>\n' }}{% else %}{{ eos_token }}{% endif %}"


# https://huggingface.co/microsoft/Phi-3-medium-4k-instruct/blob/main/tokenizer_config.json#L119
phi3_medium_template = "{% for message in messages %}{% if (message['role'] == 'user') %}{{'<|user|>' + '\n' + message['content'] + '<|end|>' + '\n' + '<|assistant|>' + '\n'}}{% elif (message['role'] == 'assistant') %}{{message['content'] + '<|end|>' + '\n'}}{% endif %}{% endfor %}"


# Although the templates are different, the roles are the same between medium and small (for now)
class Phi3SmallMediumChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = phi3_small_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<|user|>\n"
        elif role_name == "assistant":
            return "<|assistant|>\n"
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        return "<|end|>\n"


CHAT_TEMPLATE_CACHE[phi3_small_template] = Phi3SmallMediumChatTemplate
CHAT_TEMPLATE_CACHE[phi3_medium_template] = Phi3SmallMediumChatTemplate

# --------------------------------------------------
# @@@@ Mistral-7B-Instruct-v0.2 @@@@
# --------------------------------------------------
# [05/08/24] https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.2/blob/main/tokenizer_config.json#L42
mistral_7b_instruct_template = "{%- if messages[0]['role'] == 'system' %}\n    {%- set system_message = messages[0]['content'] %}\n    {%- set loop_messages = messages[1:] %}\n{%- else %}\n    {%- set loop_messages = messages %}\n{%- endif %}\n\n{{- bos_token }}\n{%- for message in loop_messages %}\n    {%- if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}\n        {{- raise_exception('After the optional system message, conversation roles must alternate user/assistant/user/assistant/...') }}\n    {%- endif %}\n    {%- if message['role'] == 'user' %}\n        {%- if loop.first and system_message is defined %}\n            {{- ' [INST] ' + system_message + '\\n\\n' + message['content'] + ' [/INST]' }}\n        {%- else %}\n            {{- ' [INST] ' + message['content'] + ' [/INST]' }}\n        {%- endif %}\n    {%- elif message['role'] == 'assistant' %}\n        {{- ' ' + message['content'] + eos_token}}\n    {%- else %}\n        {{- raise_exception('Only user and assistant roles are supported, with the exception of an initial optional system message!') }}\n    {%- endif %}\n{%- endfor %}\n"
mistral_7b_instruct_llamacpp_template = "{{ bos_token }}{% for message in messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if message['role'] == 'user' %}{{ '[INST] ' + message['content'] + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ message['content'] + eos_token}}{% else %}{{ raise_exception('Only user and assistant roles are supported!') }}{% endif %}{% endfor %}"


class Mistral7BInstructChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = mistral_7b_instruct_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return " [INST] "
        elif role_name == "assistant":
            return " "
        elif role_name == "system":
            raise ValueError("Please include system instructions in the first user message")
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return "</s>"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[mistral_7b_instruct_template] = Mistral7BInstructChatTemplate
CHAT_TEMPLATE_CACHE[mistral_7b_instruct_llamacpp_template] = Mistral7BInstructChatTemplate

# --------------------------------------------------
# @@@@ Gemma-2-9b-it @@@@
# --------------------------------------------------
# From https://huggingface.co/google/gemma-2-9b-it/blob/main/tokenizer_config.json#L1747
gemma2_9b_it_template = "{{ bos_token }}{% if messages[0]['role'] == 'system' %}{{ raise_exception('System role not supported') }}{% endif %}{% for message in messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if (message['role'] == 'assistant') %}{% set role = 'model' %}{% else %}{% set role = message['role'] %}{% endif %}{{ '<start_of_turn>' + role + '\n' + message['content'] | trim + '<end_of_turn>\n' }}{% endfor %}{% if add_generation_prompt %}{{'<start_of_turn>model\n'}}{% endif %}"


class Gemma29BInstructChatTemplate(ChatTemplate):
    # available_roles = ["user", "assistant"]
    template_str = gemma2_9b_it_template

    def get_role_start(self, role_name):
        if role_name == "user":
            return "<start_of_turn>user\n"
        elif role_name == "assistant":
            return "<start_of_turn>model\n"
        elif role_name == "system":
            raise ValueError("System Role not supported")
        else:
            raise UnsupportedRoleException(role_name, self)

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return "<end_of_turn>\n"
        elif role_name == "assistant":
            return "<end_of_turn>\n"
        else:
            raise UnsupportedRoleException(role_name, self)


CHAT_TEMPLATE_CACHE[gemma2_9b_it_template] = Gemma29BInstructChatTemplate



# mypy: ignore-errors
# NOTE(nopdive): This module has multiple imports from
#   dependencies without stubs / markers. Ignoring mypy for now.
"""Backend powerlift integration for benchmarking."""

from typing import Generator, List, Optional, Tuple, Union
import pathlib
from pathlib import Path


def retrieve_langchain(
    cache_dir: Optional[Union[str, Path]] = None,
) -> Generator[object, None, None]:
    """Retrieves LangChain datasets appropriate for guidance benchmarking. Requires env `LANGCHAIN_API_KEY` to be set on first call.

    Args:
        cache_dir (Optional[Union[str, Path]], optional): Directory to store downloaded datasets. Defaults to None.

    Yields:
        Generator[object, None, None]: DataFrameDataset required for powerlift.
    """

    import pandas as pd
    from langchain_benchmarks import registry
    from langsmith.client import Client
    from langchain_benchmarks import clone_public_dataset
    from powerlift.bench.store import update_cache, retrieve_cache, DataFrameDataset
    import os

    if cache_dir is not None:
        cache_dir = pathlib.Path(cache_dir, "langchain")

    name = "chat_extract"
    inputs_name = f"{name}.inputs.parquet"
    outputs_name = f"{name}.outputs.parquet"
    meta_name = f"{name}.meta.json"
    cached = retrieve_cache(cache_dir, [inputs_name, outputs_name, meta_name])
    if cached is None:
        if os.getenv("LANGCHAIN_API_KEY") is None:
            raise ValueError("Env variable LANGCHAIN_API_KEY is not set.")

        task = registry["Chat Extraction"]
        clone_public_dataset(task.dataset_id, dataset_name=task.name)
        client = Client()
        dataset = client.read_dataset(dataset_name=task.name)
        examples = list(client.list_examples(dataset_id=dataset.id, as_of=None))

        system_prompt = (
            "You are a data extraction bot tasked with "
            + "extracting and inferring information from dialogues and generating tickets. Always respond "
            + f"only with json based on the following JSON schema:\n{task.schema.schema_json()}"
        )

        user_prompts = []
        for example in examples:
            dialogue = (
                f"<question>\n{example.inputs['question']}\n</question>\n"
                + f"<assistant-response>\n{example.inputs['answer']}\n</assistant-response>"
            )
            user_prompt = (
                "Generate a ticket from the following question-response pair:\n"
                + f"<Dialogue>\n{dialogue}\n</Dialogue>\n"
                + "Remember, respond directly with this format:\n"
                + f'{{"{task.schema.schema()["title"]}": ...}}\n'
                + "RESPOND ONLY IN JSON THEN STOP."
            )
            user_prompts.append(user_prompt)

        inputs = pd.DataFrame.from_records(
            [
                {
                    "input": example.inputs,
                    "schema": task.schema.schema_json(),
                    "system_prompt": system_prompt,
                    "user_prompt": user_prompts[i],
                }
                for i, example in enumerate(examples)
            ]
        )

        outputs = pd.DataFrame.from_records(
            [{"output": example.outputs} for example in examples]
        )
        meta = {
            "name": name,
            "problem": "guidance/struct_decode",
            "source": "langchain",
            "inputs_categorical_mask": [dt.kind == "O" for dt in inputs.dtypes],
            "outputs_categorical_mask": [dt.kind == "O" for dt in outputs.dtypes],
            "inputs_feature_names": list(inputs.columns),
            "outputs_feature_names": list(outputs.columns),
        }
        guidance_dataset = DataFrameDataset(inputs, outputs, meta)
    else:
        guidance_dataset = DataFrameDataset.deserialize(*cached)

    if cache_dir is not None:
        serialized = DataFrameDataset.serialize(guidance_dataset)
        update_cache(cache_dir, [inputs_name, outputs_name, meta_name], serialized)

    yield guidance_dataset


def langchain_chat_extract_filter_template(models: List[str], task):
    """Template that filters tasks in benchmark down to langchain chat extract.

    Needs to be used as a partial function with only task as an argument for powerlift.

    Args:
        models (List[str]): Model names to run.
        task: Powerlift task.
    """

    if task.problem == "guidance/struct_decode":
        return models
    return []


def langchain_chat_extract_runner(trial):
    """Runs a single trial for langchain chat extract

    Args:
        trial: Powerlift trial.
    """
    from guidance import models, gen, guidance, select, zero_or_more
    from time import time
    import pandas as pd
    import json
    import json_stream
    import io
    from huggingface_hub import hf_hub_download

    if trial.task.name == "chat_extract":
        inputs, outputs, meta = trial.task.data(["inputs", "outputs", "meta"])
        merged_df = pd.concat(
            [inputs.reset_index(drop=True), outputs.reset_index(drop=True)], axis=1
        )

        if trial.method.name.startswith("guidance"):
            QUESTION_CAT = [
                "Implementation Issues",
                "Feature Requests",
                "Concept Explanations",
                "Code Optimization",
                "Security and Privacy Concerns",
                "Model Training and Fine-tuning",
                "Data Handling and Manipulation",
                "User Interaction Flow",
                "Technical Integration",
                "Error Handling and Logging",
                "Customization and Configuration",
                "External API and Data Source Integration",
                "Language and Localization",
                "Streaming and Real-time Processing",
                "Tool Development",
                "Function Calling",
                "LLM Integrations",
                "General Agent Question",
                "General Chit Chat",
                "Memory",
                "Debugging Help",
                "Application Design",
                "Prompt Templates",
                "Cost Tracking",
                "Other",
            ]
            RESPONSE_TYPE = [
                "resolve issue",
                "provide guidance",
                "request information",
                "give up",
                "none",
                "other",
            ]

            @guidance(stateless=True, dedent=False)
            def guidance_list(lm):
                return (
                    lm
                    + "["
                    + zero_or_more(gen(regex=r'"[\w ]+", '))
                    + gen(regex=r'"[\w ]+"')
                    + "]"
                )

            WORD_PAT = r"[\w ]+"
            NEW_REC = "\n            "
            DOUBLE_QUOTE = '"'
            NEW_LINE = "\n"

            @guidance(stateless=False, dedent=False)
            def gen_chat_json(lm):
                lm += f"""{{
                    "GenerateTicket": {{
                        "issue_summary": "{gen(regex=WORD_PAT, stop='"')}",
                        "question": {{
                            "question_category": "{select(QUESTION_CAT, name='question_cat')}",
                            """
                if lm["question_cat"] == "Other":
                    lm += f""""category_if_other": "{gen(regex=WORD_PAT, stop='"')}",
                            """

                lm += f""""is_off_topic": {select(["false", "true"])},
                            "toxicity": {select([0, 1, 2, 3, 4, 5])},
                            "sentiment": "{select(["Negative", "Neutral", "Positive"])}",
                            "programming_language": "{select(["python", "javascript", "typescript", "unknown", "other"])}"
                        }},
                        "response": {{
                            "response_type": "{select(RESPONSE_TYPE, name='response_type')}",
                            """

                if lm["response_type"] == "other":
                    lm += f""""response_type_if_other": "{gen(regex=WORD_PAT, stop='"')}",
                            """

                lm += f""""confidence_level": {select([0, 1, 2, 3, 4, 5])}"""
                lm += f"""{select(['', ',' + NEW_REC + '"followup_actions":'], name='follow_up')}"""
                follow_up = lm.get("follow_up", None)
                if follow_up is not None and follow_up != '':
                    lm += f""" {guidance_list()}"""

                lm += f"""
                    }}
                }}
            }}"""
                return lm

        for i, row in merged_df.iterrows():
            # Initialize LLM
            if i == 0:
                if "mistral" in trial.method.name:
                    lm_path = hf_hub_download(
                        "TheBloke/Mistral-7B-Instruct-v0.2-GGUF",
                        "mistral-7b-instruct-v0.2.Q8_0.gguf",
                    )
                elif "llama2-7b" in trial.method.name:
                    lm_path = hf_hub_download(
                        "TheBloke/Llama-2-7B-32K-Instruct-GGUF",
                        "llama-2-7b-32k-instruct.Q8_0.gguf",
                    )
                elif "phi-3" in trial.method.name:
                    lm_path = hf_hub_download(
                        "microsoft/Phi-3-mini-4k-instruct-gguf",
                        "Phi-3-mini-4k-instruct-fp16.gguf",
                    )
                else:
                    raise ValueError(
                        f"No support for method {trial.method.name}"
                    )  # pragma: no cover

                base_lm = models.LlamaCpp(
                    lm_path,
                    n_ctx=8192,
                    n_gpu_layers=-1,
                    echo=False,
                    verbose=False,
                )

            # Execute LLM
            print(f"{trial.method.name}[{i}]")
            start_time = time()
            lm = base_lm
            lm.engine.reset_metrics()
            if "mistral" in trial.method.name:
                lm += f"""<s>[INST] {row['system_prompt']}\n{row['user_prompt']} [/INST]"""
            elif "llama" in trial.method.name:
                lm += f"""<s>[INST] <<SYS>>\n{row['system_prompt']}\n<</SYS>>\n\n{row['user_prompt']}[/INST]"""
            elif "phi" in trial.method.name:
                lm += f"""<s><|user|>{row['system_prompt']}\n{row['user_prompt']}<|end|><|assistant|>"""
            else:
                raise ValueError(
                    f"Cannot support {trial.method.name} for system prompts"
                )  # pragma: no cover

            before_idx = len(str(lm))
            if "guidance" in trial.method.name:
                lm += gen_chat_json()
            else:
                lm += gen(max_tokens=1500)
            output_str = str(lm)[before_idx:]
            end_time = time()
            elapsed_time = end_time - start_time

            # Basic measures
            trial.log("output", output_str)
            trial.log("wall_time", elapsed_time)

            # Token statistics
            tm = {
                "input": lm.engine.metrics.engine_input_tokens,
                "output": lm.engine.metrics.engine_output_tokens,
                "token_count": lm.token_count,
            }
            tm["token_reduction"] = 1 - (tm["output"]) / (lm.token_count)
            trial.log("token_input", tm["input"])
            trial.log("token_output", tm["output"])
            trial.log("token_count", tm["token_count"])
            trial.log("token_reduction", tm["token_reduction"])

            # Validate JSON conformance
            json_success = False
            try:
                output_json = json.loads(output_str.strip())
                output_json = output_json["GenerateTicket"]
                json_success = True
            except Exception as e:
                trial.log("json_errmsg", str(e))
            trial.log("json", json_success * 1)

            if json_success:
                trial.log("output_json", output_json)
                trial.log("json_dirty", 0)
            else:
                success = False
                candidate = output_str.strip()
                for i, ch in enumerate(candidate):
                    if ch == "{":
                        try:
                            results = json_stream.load(io.StringIO(candidate[i:]))
                            di = json_stream.to_standard_types(results)
                            di = di["GenerateTicket"]
                            success = True
                            break
                        except Exception:
                            pass
                if success:
                    output_json = di
                    trial.log("output_json", output_json)
                    trial.log("json_dirty", 1)
                else:
                    trial.log("output_json", {})
                    trial.log("json_dirty", 0)

            # Validate JSON schema conformance
            from langchain_benchmarks.extraction.tasks.chat_extraction.schema import (
                GenerateTicket,
            )

            try:
                GenerateTicket.parse_obj(output_json)
                trial.log("json_valid", 1)
                if json_success:
                    trial.log("json_valid_strict", 1)
                else:
                    trial.log("json_valid_strict", 0)
            except Exception as e:
                trial.log("json_valid", 0)
                trial.log("json_valid_strict", 0)
                trial.log("json_valid_errmsg", str(e))

            # Toxicity similarity
            expected_json = row["output"]["output"]
            expected = expected_json["question"]["toxicity"]
            try:
                pred = output_json["question"]["toxicity"]
                score = 1 - abs(expected - float(pred)) / 5
                trial.log("toxicity", score)
                trial.log("toxicity_strict", score)
            except Exception as e:
                trial.log("toxicity_strict", 0)
                trial.log("toxicity_errmsg", str(e))

            # Sentiment similarity
            expected = expected_json["question"]["sentiment"]
            ordinal_map = {
                "negative": 0,
                "neutral": 1,
                "positive": 2,
            }
            expected_score = ordinal_map.get(str(expected).lower())
            try:
                pred = output_json["question"]["sentiment"]
                pred_score = ordinal_map.get(str(pred).lower())
                score = 1 - (abs(expected_score - float(pred_score)) / 2)
                trial.log("sentiment", score)
                trial.log("sentiment_strict", score)
            except Exception as e:
                trial.log("sentiment_strict", 0)
                trial.log("sentiment_errmsg", str(e))

            # Question category similarity
            expected = expected_json["question"]["question_category"]
            try:
                pred = output_json["question"]["question_category"]
                score = int(expected == pred)
                trial.log("question_cat", score)
                trial.log("question_cat_strict", score)
            except Exception as e:
                trial.log("question_cat_strict", 0)
                trial.log("question_cat_errmsg", str(e))

            # Off-topic similarity
            expected = expected_json["question"]["is_off_topic"]
            try:
                pred = output_json["question"].get("is_off_topic")
                score = int(expected == pred)
                trial.log("offtopic", score)
                trial.log("offtopic_strict", score)
            except Exception as e:
                trial.log("offtopic_strict", 0)
                trial.log("offtopic_errmsg", str(e))

            # Programming language similarity
            expected = expected_json["question"]["programming_language"]
            try:
                pred = output_json["question"]["programming_language"]
                score = int(expected == pred)
                trial.log("programming", score)
                trial.log("programming_strict", score)
            except Exception as e:
                trial.log("programming_strict", 0)
                trial.log("programming_errmsg", str(e))


def bench(
    db_url: str,
    experiment_name: str,
    models: List[str],
    force_recreate: bool,
    timeout: int,
    cache_dir: Union[str, Path],
    debug_mode: bool,
) -> Tuple[object, object]:
    """Runs the benchmark.

    Requires LANGCHAIN_API_KEY to be set as an environment variable the first time.

    Args:
        db_url (str): Database connection string.
        experiment_name (str): Name of experiment to create / run.
        models (List[str]): Models to benchmark.
        force_recreate (bool): Recreate the database before benchmarking.
        timeout (int): Max execution time per trial.
        cache_dir (Union[str, Path]): Cache to store external datasets.
        debug_mode (bool): Set this when you require a debugger to step line by line in the trial runner.

    Returns:
        Tuple[object, object]: (status, results) data frames where status relates to trials, results are wide form aggregates of each model.
    """

    from powerlift.bench import Benchmark, Store, populate_with_datasets
    from powerlift.executors import LocalMachine

    store = Store(db_url, force_recreate=force_recreate)
    populate_with_datasets(
        store, retrieve_langchain(cache_dir=cache_dir), exist_ok=True
    )
    executor = LocalMachine(store, n_cpus=1, debug_mode=debug_mode)

    bench = Benchmark(store, name=experiment_name)
    bench.run(
        langchain_chat_extract_runner,
        lambda x: langchain_chat_extract_filter_template(models, x),
        timeout=timeout,
        executor=executor,
    )

    bench.wait_until_complete()
    status_df = bench.status()
    result_df = bench.results()
    agg_df = result_df.pivot_table(
        index="method", columns="name", values="num_val", aggfunc=["mean", "std"]
    )
    agg_df.columns = ["_".join(x) for x in agg_df.columns.to_flat_index()]

    return status_df, agg_df



import os
import tiktoken

from ._model import Chat, Instruct
from ._grammarless import GrammarlessEngine, Grammarless


class AnthropicEngine(GrammarlessEngine):
    def __init__(
        self,
        model,
        tokenizer,
        api_key,
        timeout,
        max_streaming_tokens,
        compute_log_probs,
        **kwargs,
    ):
        try:
            from anthropic import Anthropic
        except ModuleNotFoundError:
            raise Exception(
                "Please install the anthropic package version >= 0.7 using `pip install anthropic -U` in order to use guidance.models.Anthropic!"
            )

        if api_key is None:
            api_key = os.environ.get("ANTHROPIC_API_KEY")

        if api_key is None:
            raise Exception(
                "Expected an api_key argument or the ANTHROPIC_API_KEY environment variable to be set!"
            )

        self.anthropic = Anthropic(api_key=api_key, **kwargs)
        self.model_name = model

        # we pretend it tokenizes like gpt2 if tiktoken does not know about it... TODO: make this better
        if tokenizer is None:
            try:
                tokenizer = tiktoken.encoding_for_model(model)
            except:
                tokenizer = tiktoken.get_encoding("gpt2")

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)

    def _generator(self, prompt, temperature):

        # find the role tags
        pos = 0
        role_end = b"<|im_end|>\n"
        messages = []
        found = True
        system_prompt = None # Not mandatory, but we'll store it if found
        while found:

            # find the role text blocks
            found = False
            for role_name, start_bytes in (
                ("system", b"<|im_start|>system\n"),
                ("user", b"<|im_start|>user\n"),
                ("assistant", b"<|im_start|>assistant\n"),
            ):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert (
                            role_name == "assistant"
                        ), "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos : pos + end_pos]
                    pos += end_pos + len(role_end)
                    if role_name == "system":
                        system_prompt = btext.decode("utf8")
                    else:
                        messages.append(
                            {"role": role_name, "content": btext.decode("utf8")}
                        )
                    found = True
                    break

        # Add nice exception if no role tags were used in the prompt.
        # TODO: Move this somewhere more general for all chat models?
        if messages == []:
            raise ValueError(
                f"The AnthropicAI model {self.model_name} is a Chat-based model and requires role tags in the prompt! \
            Make sure you are using guidance context managers like `with system():`, `with user():` and `with assistant():` \
            to appropriately format your guidance program for this type of model."
            )

        # update our shared data state
        self._reset_shared_data(prompt, temperature)

        # API call and response handling
        try:
            # Need to do this because Anthropic API is a bit weird with the system keyword...
            model_kwargs = dict(
                model=self.model_name,
                messages=messages,
                max_tokens=self.max_streaming_tokens,
                temperature=temperature,
            )
            if system_prompt is not None:
                model_kwargs["system"] = system_prompt
            generator = self.anthropic.messages.stream(
                **model_kwargs,
            )
        except Exception as e:  # TODO: add retry logic
            raise e

        with generator as stream:
            for chunk in stream.text_stream:
                # print(chunk)
                yield chunk.encode("utf8")


class Anthropic(Grammarless):
    """Represents an Anthropic model as exposed through their remote API.

    Note that because this uses a remote API endpoint without built-in guidance support
    there are some things we cannot do, like force the model to follow a pattern inside
    a chat role block.
    """

    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        timeout=0.5,
        max_streaming_tokens=1000,
        compute_log_probs=False,
        **kwargs,
    ):
        """Build a new Anthropic model object that represents a model in a given state."""

        super().__init__(
            engine=AnthropicEngine(
                model=model,
                tokenizer=tokenizer,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                **kwargs,
            ),
            echo=echo,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: _serialization.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14_serialization.proto\x12\x08guidance\"3\n\x07Grammar\x12(\n\x05nodes\x18\x01 \x03(\x0b\x32\x19.guidance.GrammarFunction\"\xa5\x03\n\x12\x45ngineCallResponse\x12\x11\n\tnew_bytes\x18\x01 \x01(\x0c\x12\x14\n\x0cis_generated\x18\x02 \x01(\x08\x12\x16\n\x0enew_bytes_prob\x18\x03 \x01(\x02\x12G\n\x0e\x63\x61pture_groups\x18\x04 \x03(\x0b\x32/.guidance.EngineCallResponse.CaptureGroupsEntry\x12W\n\x17\x63\x61pture_group_log_probs\x18\x05 \x03(\x0b\x32\x36.guidance.EngineCallResponse.CaptureGroupLogProbsEntry\x12\x17\n\x0fnew_token_count\x18\x06 \x01(\x05\x1a\x45\n\x12\x43\x61ptureGroupsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0f.guidance.Value:\x02\x38\x01\x1aL\n\x19\x43\x61ptureGroupLogProbsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0f.guidance.Value:\x02\x38\x01\"\x80\x01\n\x05Value\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x15\n\x0b\x62ytes_value\x18\x02 \x01(\x0cH\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x12)\n\nlist_value\x18\x04 \x01(\x0b\x32\x13.guidance.ListValueH\x00\x42\x06\n\x04kind\",\n\tListValue\x12\x1f\n\x06values\x18\x01 \x03(\x0b\x32\x0f.guidance.Value\"w\n\x04\x42yte\x12\x0c\n\x04\x62yte\x18\x01 \x01(\x0c\x12\x0e\n\x06hidden\x18\x02 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x03 \x01(\x08\x12\x10\n\x08nullable\x18\x04 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x05 \x01(\t\x12\x13\n\x0btemperature\x18\x06 \x01(\x02\"p\n\tByteRange\x12\x12\n\nbyte_range\x18\x01 \x01(\x0c\x12\x0e\n\x06hidden\x18\x03 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x04 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x05 \x01(\t\x12\x13\n\x0btemperature\x18\x06 \x01(\x02\"\x06\n\x04Null\"k\n\rModelVariable\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06hidden\x18\x02 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x03 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x04 \x01(\t\x12\x10\n\x08nullable\x18\x05 \x01(\x08\"\x86\x01\n\x04Join\x12\x10\n\x08nullable\x18\x01 \x01(\x08\x12\x0e\n\x06values\x18\x02 \x03(\x05\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06hidden\x18\x04 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x05 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x06 \x01(\t\x12\x12\n\nmax_tokens\x18\x07 \x01(\x05\"\x9b\x01\n\x06Select\x12\x10\n\x08nullable\x18\x01 \x01(\x08\x12\x0e\n\x06values\x18\x02 \x03(\x05\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0e\n\x06hidden\x18\x04 \x01(\x08\x12\x14\n\x0c\x63ommit_point\x18\x05 \x01(\x08\x12\x14\n\x0c\x63\x61pture_name\x18\x06 \x01(\t\x12\x12\n\nmax_tokens\x18\x07 \x01(\x05\x12\x11\n\trecursive\x18\x08 \x01(\x08\"\xe4\x01\n\x0fGrammarFunction\x12\x1e\n\x04join\x18\x01 \x01(\x0b\x32\x0e.guidance.JoinH\x00\x12\"\n\x06select\x18\x02 \x01(\x0b\x32\x10.guidance.SelectH\x00\x12\x1e\n\x04\x62yte\x18\x03 \x01(\x0b\x32\x0e.guidance.ByteH\x00\x12)\n\nbyte_range\x18\x04 \x01(\x0b\x32\x13.guidance.ByteRangeH\x00\x12\x31\n\x0emodel_variable\x18\x05 \x01(\x0b\x32\x17.guidance.ModelVariableH\x00\x42\x0f\n\rfunction_typeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, '_serialization_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY']._options = None
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY']._serialized_options = b'8\001'
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY']._options = None
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY']._serialized_options = b'8\001'
  _globals['_GRAMMAR']._serialized_start=34
  _globals['_GRAMMAR']._serialized_end=85
  _globals['_ENGINECALLRESPONSE']._serialized_start=88
  _globals['_ENGINECALLRESPONSE']._serialized_end=509
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY']._serialized_start=362
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPSENTRY']._serialized_end=431
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY']._serialized_start=433
  _globals['_ENGINECALLRESPONSE_CAPTUREGROUPLOGPROBSENTRY']._serialized_end=509
  _globals['_VALUE']._serialized_start=512
  _globals['_VALUE']._serialized_end=640
  _globals['_LISTVALUE']._serialized_start=642
  _globals['_LISTVALUE']._serialized_end=686
  _globals['_BYTE']._serialized_start=688
  _globals['_BYTE']._serialized_end=807
  _globals['_BYTERANGE']._serialized_start=809
  _globals['_BYTERANGE']._serialized_end=921
  _globals['_NULL']._serialized_start=923
  _globals['_NULL']._serialized_end=929
  _globals['_MODELVARIABLE']._serialized_start=931
  _globals['_MODELVARIABLE']._serialized_end=1038
  _globals['_JOIN']._serialized_start=1041
  _globals['_JOIN']._serialized_end=1175
  _globals['_SELECT']._serialized_start=1178
  _globals['_SELECT']._serialized_end=1333
  _globals['_GRAMMARFUNCTION']._serialized_start=1336
  _globals['_GRAMMARFUNCTION']._serialized_end=1564
# @@protoc_insertion_point(module_scope)



import numpy as np
from ._tokenizer import Tokenizer
from ..chat import load_template_class
import typing

class ByteTokenizer(Tokenizer):
    def __init__(self, chat_template=None):
        # directly map integer values to byte strings
        tokens = np.array([bytes([i]) for i in range(256)], dtype="object")
        chat_template = load_template_class(chat_template)
        super().__init__(tokens, chat_template)

    def __call__(self, byte_string) -> typing.List[int]:
        """Returns a list of tokens that represent the given byte string."""
        return list(byte_string)



"""User facing API for benchmarking."""

from typing import List, Optional, Tuple, Union
from pathlib import Path
from guidance.bench._utils import lib_bench_dir

"""Available models to run benchmark against."""
AVAILABLE_MODELS = [
    "guidance-mistral-7b-instruct",
    "base-mistral-7b-instruct",
    "guidance-phi-3-mini-4k-instruct",
    "base-phi-3-mini-4k-instruct",
    "guidance-llama2-7b-32k-instruct",
    "base-llama2-7b-32k-instruct",
]


def bench(
    db_url: str,
    experiment_name: str,
    models: List[str] = AVAILABLE_MODELS,
    force_recreate: bool = False,
    timeout: int = 3600,
    cache_dir: Union[str, Path] = lib_bench_dir() / "cache",
    debug_mode: bool = False,
) -> Tuple[object, object]:
    """Benchmarks guidance against preset tasks.

    This runs on a single machine, one trial at a time.
    To run this the first time you will need API_LANGCHAIN_KEY set as an environment variable.

    Args:
        db_url (str): Database connection string.
        experiment_name (str): Name of experiment to create / run.
        models (List[str], optional): Models to benchmark. Defaults to AVAILABLE_MODELS.
        force_recreate (bool, optional): Recreate the database before benchmarking. Defaults to False.
        timeout (int, optional): Max execution time per trial. Defaults to 3600.
        cache_dir (Union[str, Path], optional): Cache to store external datasets. Defaults to lib_bench_dir() / "cache".
        debug_mode (bool): Set this when you require a debugger to step line by line in the trial_runner.

    Returns:
        Tuple[object, object]: (status, results) data frames where status relates to trials, results are wide form aggregates of each model.
    """
    from guidance.bench._powerlift import bench as inner_bench

    status_df, result_df = inner_bench(
        db_url, experiment_name, models, force_recreate, timeout, cache_dir, debug_mode
    )
    return status_df, result_df



from .._guidance import guidance
from ..models._model import Model
from ._block import block


@guidance
def set_attr_opener(lm, name, value, cache):
    if hasattr(lm, name):
        cache[name] = getattr(lm, name)
    return lm.setattr(name, value)


@guidance
def set_attr_closer(lm, name, cache):
    if not Model._grammar_only:
        if name in cache:
            return lm.setattr(name, cache[name])
        else:
            return lm.delattr(name)
    else:
        return lm


def set_attribute(name, value=True):
    cache = {}
    return block(
        opener=set_attr_opener(name, value, cache),
        closer=set_attr_closer(name, cache),
    )



import requests
import os
import base64
import json
import urllib.parse
from ._model import Engine, Model, EngineCallResponse
from ..chat import Phi3MiniChatTemplate
from ._byte_tokenizer import ByteTokenizer


class AzureGuidanceEngine(Engine):
    """This connects to a remote guidance server on Azure and runs all computation using the remote engine."""

    def __init__(self, server_url, max_streaming_tokens=1000, chat_template=None):
        if (
            server_url is None
            or isinstance(server_url, str)
            and len(server_url.strip()) == 0
        ):
            server_url = os.getenv("AZURE_GUIDANCE_URL", "")
        elif not isinstance(server_url, str):
            raise ValueError("server_url must contain a URL string.")

        if not server_url.startswith("https://"):
            raise ValueError(
                "AzureGuidance requires a remote model URL that starts with https://"
            )
        self.server_url = server_url
        self.max_streaming_tokens = max_streaming_tokens

        if chat_template is None:
            # TODO [PK]: obtain this from the server
            chat_template=Phi3MiniChatTemplate

        tokenizer = ByteTokenizer(chat_template)

        # build the Engine
        super().__init__(tokenizer=tokenizer, compute_log_probs=False)

    def __call__(self, parser, grammar, ensure_bos_token=True):
        b64 = base64.b64encode(grammar.serialize()).decode("utf-8")

        data = {
            "controller": "guidance",
            "controller_arg": {"guidance_b64": b64},
            "prompt": parser,
            "max_tokens": self.max_streaming_tokens,
            "temperature": 0.0, # this is just default temperature
        }

        resp = req("post", "run", json=data, stream=True, base_url=self.server_url)
        if resp.status_code != 200:
            text = resp.text
            try:
                d = json.loads(text)
                if "message" in d:
                    text = d["message"]
            except:
                pass
            raise RuntimeError(
                f"Bad response to Guidance request {resp.status_code} {resp.reason}: {text}."
            )

        for line in resp.iter_lines():
            if not line:
                continue
            decoded_line: str = line.decode("utf-8")
            if decoded_line.startswith("data: {"):
                d = json.loads(decoded_line[6:])
                if "forks" not in d:
                    continue
                for ch in d["forks"]:
                    capture_groups = {}
                    capture_group_log_probs = {}

                    if "Previous WASM Error" in ch["logs"]:
                        raise RuntimeError("Previous WASM Error.")
                    idx = ch["index"]
                    assert idx == 0, "unexpected index in response from server"
                    new_bytes = b""
                    new_token_count = 0
                    new_bytes_prob = 0.0
                    num_text_entries = 0
                    for ln in ch["logs"].split("\n"):
                        ln: str
                        if ln.startswith("JSON-OUT: "):
                            j = json.loads(ln[10:])
                            tag = j.get("object", "")
                            if tag == "capture":
                                capture_groups[j["name"]] = bytes.fromhex(j["hex"])
                                capture_group_log_probs[j["name"]] = j["log_prob"]
                            elif tag == "text":
                                # it actually should only happen once per round...
                                new_bytes += bytes.fromhex(j["hex"])
                                new_token_count += j["num_tokens"]
                                new_bytes_prob += j["log_prob"]
                                num_text_entries += 1
                    if num_text_entries > 0:
                        new_bytes_prob /= num_text_entries

                    # print(ch["logs"].rstrip("\n"), flush=True)

                    err = ch.get("error", "")
                    if err:
                        raise RuntimeError(f"Error returned by grammar server {err}.")

                    is_generated = True  # TODO: get this from the server

                    response_data = EngineCallResponse(
                        new_bytes,
                        is_generated,
                        new_bytes_prob,
                        capture_groups,
                        capture_group_log_probs,
                        new_token_count,
                    )
                    yield response_data
            elif decoded_line == "data: [DONE]":
                pass
            else:
                raise RuntimeError(f"bad response line: {decoded_line}")


class AzureGuidance(Model):
    def __init__(
        self,
        model=None,
        echo=True,
        max_streaming_tokens=1000,
        chat_template=None,
    ):
        """Build a new remote grammar processing Azure model object that represents a model in a given state."""

        engine = AzureGuidanceEngine(model, max_streaming_tokens, chat_template)
        super().__init__(engine, echo=echo)


def _parse_base_url(base_url: str):
    p = urllib.parse.urlparse(base_url)
    key = ""
    if p.fragment:
        f = urllib.parse.parse_qs(p.fragment)
        key = f.get("key", [""])[0]
    r = urllib.parse.urlunparse(p._replace(fragment="", query=""))
    if not r.endswith("/"):
        r += "/"
    return r, key


def _headers(arg_base_url: str) -> dict:
    _, key = _parse_base_url(arg_base_url)
    if key:
        return {"api-key": key}
    else:
        return {}


def _mk_url(path: str, arg_base_url: str) -> str:
    pref, _ = _parse_base_url(arg_base_url)
    return pref + path


def req(tp: str, path: str, base_url: str, **kwargs):
    url = _mk_url(path, arg_base_url=base_url)
    headers = _headers(arg_base_url=base_url)
    resp = requests.request(tp, url, headers=headers, **kwargs)
    return resp



import os
import re
import textwrap
import warnings

from typing import Sequence, Union

try:
    import torch
except ModuleNotFoundError:
    pass


try:
    import transformers as transformers_package

    has_transformers = True
except ModuleNotFoundError:
    has_transformers = False


from .._model import Engine, Model
from .._tokenizer import Tokenizer

# Formed by comparing model and tokenizer from_pretrained methods
# transformers/models/auto/auto_factory.py
# transformers/models/auto/tokenization_auto.py
_COMMON_TRANSFORMERS_KWARGS = [
    "cache_dir",
    "force_download",
    "proxies",
    "resume_download",
    "revision",
    "subfolder",
    "trust_remote_code",
]

class ByteDecoderError(Exception):
    pass

class ByteTokensError(Exception):
    pass

class TransformersTokenizer(Tokenizer):
    def __init__(
        self,
        model: Union[str, "transformers_package.PreTrainedModel"],
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
            None,
        ],
        chat_template=None,
        ignore_bos_token=False,
        **kwargs,
    ):
        if transformers_tokenizer is None:
            if isinstance(model, str):
                transformers_tokenizer, byte_tokens = self._tokenizer(model, **kwargs)
            else:
                raise ValueError(
                    "A model object was passed in, but no tokenizer was provided. Please provide a tokenizer."
                )
        else:
            is_ptt = isinstance(transformers_tokenizer, transformers_package.PreTrainedTokenizer)
            is_ptt_fast = isinstance(
                transformers_tokenizer, transformers_package.PreTrainedTokenizerFast
            )
            assert is_ptt or is_ptt_fast
            byte_tokens = self._byte_tokens(transformers_tokenizer)

        self._orig_tokenizer = transformers_tokenizer

        # Chat Template logic
        if chat_template is None and hasattr(self._orig_tokenizer, "chat_template"):
            chat_template = self._orig_tokenizer.chat_template

        # the superclass does most of the work once we have the tokens
        super().__init__(
            byte_tokens,
            chat_template,
            None if ignore_bos_token else transformers_tokenizer.bos_token_id,
            transformers_tokenizer.eos_token_id,
        )

    def _tokenizer(self, model: str, **kwargs) -> tuple[
        Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
        list[bytes],
    ]:
        # make sure transformers is installed
        if not has_transformers:
            raise ImportError("Please install transformers with `pip install transformers`")

        try:
            tokenizer = transformers_package.AutoTokenizer.from_pretrained(
                model, use_fast=False, **kwargs
            )
            byte_tokens = self._byte_tokens(tokenizer)
        except ImportError:
            # Raise on ImportError because it's likely a missing dependency that the user can install
            raise
        except ByteTokensError as e:
            # Give a specific warning for ByteTokensError and fall back to fast tokenizer
            warnings.warn(f"Falling back to fast tokenizer. Could not build byte tokens for model {model!r} due to exception {e.__class__.__name__}: {e}")
        except Exception as e:
            # Fall back for other exceptions
            warnings.warn(f"Falling back to fast tokenizer. Could not load tokenizer for model {model!r} due to exception {e.__class__.__name__}: {e}")
        else:
            return tokenizer, byte_tokens

        tokenizer = transformers_package.AutoTokenizer.from_pretrained(
            model, use_fast=True, **kwargs
        )
        try:
            byte_tokens = self._byte_tokens(tokenizer)
        except ByteTokensError as e:
            raise ValueError(f"Fallback to fast tokenizer failed for model {model!r}") from e
        return tokenizer, byte_tokens

    def _byte_tokens(
        self,
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
    ) -> list[bytes]:

        if hasattr(transformers_tokenizer, "byte_decoder"):
            try:
                self._check_byte_decoder(
                    transformers_tokenizer.byte_decoder, transformers_tokenizer
                )
            except ByteDecoderError as e:
                warnings.warn(
                    f"Tokenizer has a byte_decoder, but it can't be used to construct byte_tokens: {e}"
                )
                pass
            else:
                return self._byte_tokens_from_byte_decoder(transformers_tokenizer.byte_decoder, transformers_tokenizer)

        if hasattr(transformers_tokenizer, "sp_model"):
            return self._byte_tokens_from_sp_model(transformers_tokenizer)

        try:
            return self._byte_tokens_by_encoding_token_strings(transformers_tokenizer)
        except ValueError as e:
            warnings.warn(
                f"Could not build_byte tokens from the tokenizer by encoding token strings: {e}"
            )
            pass

        fallback_byte_decoder = self._fallback_byte_decoder()
        try:
            self._check_byte_decoder(
                fallback_byte_decoder, transformers_tokenizer
            )
        except ByteDecoderError as e:
            # Should be the only exception that is raised in _byte_tokens
            raise ByteTokensError(
                "Could not build byte tokens from the tokenizer, and falling back to a standard gpt2 byte_decoder failed"
            ) from e
        return self._byte_tokens_from_byte_decoder(fallback_byte_decoder, transformers_tokenizer)

    def _byte_tokens_from_byte_decoder(
        self,
        byte_decoder: dict[str, int],
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
    ) -> list[bytes]:
        byte_tokens = [b""] * len(transformers_tokenizer)
        for i in range(len(transformers_tokenizer)):
            byte_coded = bytes(
                [byte_decoder[c] for c in transformers_tokenizer.convert_ids_to_tokens(i)]
            )
            byte_tokens[i] = byte_coded
        return byte_tokens

    def _byte_tokens_from_sp_model(
        self,
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
    ) -> list[bytes]:
        byte_tokens = [b""] * len(transformers_tokenizer)
        special_tokens_map = {
            id: token for token, id in transformers_tokenizer.get_added_vocab().items()
        }
        space_prefix = "▁".encode()
        for i in range(len(transformers_tokenizer)):
            if i in special_tokens_map:
                byte_coded = special_tokens_map[i].encode()
            else:
                byte_coded = re.sub(
                    rb"<0x(..)>",
                    lambda x: bytes.fromhex(x[1].decode()),
                    transformers_tokenizer.sp_model.id_to_piece(i).encode(),
                )
            byte_tokens[i] = byte_coded.replace(space_prefix, b" ")
        return byte_tokens

    def _byte_tokens_by_encoding_token_strings(
        self,
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
    ) -> list[bytes]:
        byte_tokens = [b""] * len(transformers_tokenizer)
        special_tokens_map = {
            id: token for token, id in transformers_tokenizer.get_added_vocab().items()
        }
        byte_encoder = self._bytes_to_unicode()
        byte_decoder = {v: k for k, v in byte_encoder.items()}

        for i in range(len(transformers_tokenizer)):
            if i in special_tokens_map:
                byte_coded = special_tokens_map[i].encode()
            else:
                token = transformers_tokenizer.convert_ids_to_tokens(i)
                if isinstance(token, bytes):
                    byte_coded = token
                elif isinstance(token, str):
                    if hasattr(transformers_tokenizer, "convert_tokens_to_string"):
                        token_str = transformers_tokenizer.convert_tokens_to_string([token])
                        encoded_str = transformers_tokenizer.encode(token_str)
                        if len(encoded_str) != 1:
                            raise ValueError(f"Round-trip encoding of tokens [{token}] failed! Got {encoded_str}")
                        roundtrip_id = encoded_str[0]
                        if roundtrip_id == i:
                            byte_coded = token_str.encode()
                        else:
                            byte_coded = bytes([byte_decoder[c] for c in token])
                    else:
                        byte_coded = token.encode()
                else:
                    raise ValueError(f"Unexpected token type: {type(token)}")
            byte_tokens[i] = byte_coded
        return byte_tokens

    def _fallback_byte_decoder(self) -> dict[str, int]:
        byte_decoder = transformers_package.AutoTokenizer.from_pretrained(
            "gpt2", use_fast=False
        ).byte_decoder # fall back to gpt2 mapping

        # some special tokens may not have their whitespace encoded...
        byte_decoder[" "] = 32
        byte_decoder["\n"] = 10
        byte_decoder["\r"] = 13
        byte_decoder["\t"] = 9
        byte_decoder["▁"] = 32

        return byte_decoder

    def _check_byte_decoder(
        self,
        byte_decoder: dict[str, int],
        transformers_tokenizer: Union[
            "transformers_package.PreTrainedTokenizer",
            "transformers_package.PreTrainedTokenizerFast",
        ],
    ) -> None:

        def check_byte_decoder_has_all_bytes() -> None:
            # This is here because some tokenizers are bad and don't have all the bytes (I'm looking at you, microsoft/phi2)
            all_bytes = set()
            for x in transformers_tokenizer.get_vocab().keys():
                for y in x:
                    all_bytes.add(y)
            if not set(byte_decoder.keys()) >= all_bytes:
                raise ByteDecoderError(
                    f"Byte decoder is missing bytes: {all_bytes - set(byte_decoder.keys())}"
                )

        def check_byte_decoder_complex_round_trip() -> None:
            # run a quick spot check to verify we can rebuild complex multi-token unicode symbols
            s = "’•¶∂ƒ˙∆£Ħ爨ൠᅘ∰፨"
            reconstructed = b""
            try:
                input_ids = transformers_tokenizer(s)["input_ids"]
                for i in input_ids:
                    nxt_bytes = []
                    token_str = transformers_tokenizer.convert_ids_to_tokens(i)
                    for c in token_str:
                        nxt_bytes.append(byte_decoder[c])
                    reconstructed += bytes(nxt_bytes)
                # Check if the tokenizer has a bos_token attribute, and if it does, check
                # if it's at the start of the reconstructed bytes
                # Some tokenizers add this automatically as part of the call function, so
                # we need to remove it to compare
                if hasattr(transformers_tokenizer, "bos_token") and transformers_tokenizer.bos_token and reconstructed.startswith(
                    transformers_tokenizer.bos_token.encode()
                ):
                    reconstructed = reconstructed[len(transformers_tokenizer.bos_token) :]
            # TODO: can we narrow this exception?
            except Exception as e:
                msg = textwrap.dedent(
                    f"""
                    The tokenizer being used is unable to convert a special character in {s}.
                    For models with sentencepiece based tokenizers (e.g. llama, phi-3-mini),
                    installing sentencepiece often fixes this issue (pip install sentencepiece).
                    """
                )
                raise ByteDecoderError(msg) from e
            if reconstructed.decode() != s:
                raise ByteDecoderError(
                    f"Failed to reconstruct the string {s} from the tokenizer's byte_decoder: {reconstructed.decode()!r} != {s!r}"
                )

        check_byte_decoder_has_all_bytes()
        check_byte_decoder_complex_round_trip()

    def _bytes_to_unicode(self):
        bs = (
            list(range(ord("!"), ord("~") + 1))
            + list(range(ord("¡"), ord("¬") + 1))
            + list(range(ord("®"), ord("ÿ") + 1))
        )
        cs = bs[:]
        n = 0
        for b in range(256):
            if b not in bs:
                bs.append(b)
                cs.append(256 + n)
                n += 1
        cs = [chr(n) for n in cs]
        return dict(zip(bs, cs))

    def encode(self, byte_string: bytes) -> Sequence[int]:
        assert isinstance(byte_string, bytes)
        # HF tokenizers take in strings apparently
        tokenization = self._orig_tokenizer(byte_string.decode(), add_special_tokens=False)
        return tokenization["input_ids"]

    def decode(self, tokens: Sequence[int]) -> bytes:
        decoded_str = self._orig_tokenizer.decode(tokens)
        return decoded_str.encode()

    def recode(self, tokens: Sequence[int]) -> Sequence[int]:
        # the encode/decode cycle might not work if we have partial unicode strings
        used_tokens = len(tokens)
        for _ in range(3):
            try:
                first_decode = self.decode(tokens).decode("utf8")
            except UnicodeDecodeError:
                if used_tokens == 0:
                    break
                else:
                    used_tokens -= 1

        new_ids = list(self.encode(first_decode.encode("utf-8")))
        if used_tokens < len(tokens):
            new_ids += tokens[used_tokens:]

        # HACK: check for a bug in the HuggingFace tokenizer
        # (that will just add extra spaces during an encode-decode cycle)
        second_decode = self._orig_tokenizer.decode(new_ids)
        if (
            second_decode != first_decode
            and len(second_decode) == len(first_decode) + 1
            and second_decode.startswith("<s>  ")
        ):
            new_ids = new_ids[0:1] + new_ids[2:]

        return new_ids


class TransformersEngine(Engine):
    def __init__(self, model, tokenizer, compute_log_probs: bool, chat_template=None, **kwargs):
        # fill in default model value
        if model is None:
            model = os.environ.get("TRANSFORMERS_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser("~/.transformers_model"), "r") as file:
                    model = file.read().replace("\n", "")
            except:
                pass

        self.model_obj = self._model(model, **kwargs)

        if not isinstance(model, str):
            self.model = model.__class__.__name__
        self.device = self.model_obj.device  # otherwise note the current device

        self._past_key_values = None
        self._cached_logits = None
        self._cached_token_ids: list[int] = []

        # Set attr for malformed tokenizer hack.
        # If more models start doing this, generalize into a util function.
        if hasattr(self.model_obj.config, "model_type"):
            if self.model_obj.config.model_type in ["phi3"]:
                self._disable_retokenize_check = True

        # Automatically fill common args between Transformers
        # model and tokenizer
        passed_common_kwargs = {}
        for arg_name in _COMMON_TRANSFORMERS_KWARGS:
            if arg_name in kwargs:
                passed_common_kwargs[arg_name] = kwargs[arg_name]

        # Create the tokenizer
        my_tokenizer = TransformersTokenizer(
            model, tokenizer, chat_template, **passed_common_kwargs
        )

        super().__init__(
            my_tokenizer,
            compute_log_probs=compute_log_probs,
        )
        assert self._token_trie.match

    def _model(self, model, **kwargs):
        # intantiate the model if needed
        if isinstance(model, str):

            # make sure transformers is installed
            if not has_transformers:
                raise Exception(
                    "Please install transformers with `pip install transformers` in order to use guidance.models.Transformers!"
                )
            model = transformers_package.AutoModelForCausalLM.from_pretrained(model, **kwargs)
        return model

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """Computes the logits for the given token state.

        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        """

        # make sure we don't run off the end of the model
        if len(token_ids) >= getattr(self.model_obj.config, "max_position_embeddings", 1e10):
            raise Exception(
                f"Attempted to run a transformers model past its maximum context window size of {self.model_obj.config.max_position_embeddings}!"
            )

        # get the number of cache positions we are using
        cache_token_ids = self._cached_token_ids
        num_cached = 0
        for id in cache_token_ids:
            if (
                num_cached >= len(cache_token_ids)
                or num_cached >= len(token_ids)
                or token_ids[num_cached] != id
            ):
                break
            num_cached += 1

        # reset the cache length according to that number of positions
        past_key_values = self._past_key_values
        past_length = past_key_values[0][0].size(-2) if past_key_values is not None else 0
        if past_length > num_cached:
            # note we recompute the last token because we don't bother to handle the special case of just computing logits
            past_length = max(0, num_cached - 1)
            self._past_key_values = tuple(
                tuple(p[..., :past_length, :] for p in v) for v in past_key_values
            )
        cache_token_ids[past_length:] = []

        # call the model
        new_token_ids = token_ids[past_length:]
        if len(new_token_ids) > 0:
            with torch.no_grad():
                # Not all models support batched tokens for some reason
                try:
                    model_out = self.model_obj(
                        input_ids=torch.tensor(new_token_ids).unsqueeze(0).to(self.device),
                        past_key_values=self._past_key_values,
                        use_cache=True,
                        position_ids=torch.arange(past_length, past_length + len(new_token_ids))
                        .unsqueeze(0)
                        .to(self.device),
                        attention_mask=torch.ones(1, past_length + len(new_token_ids)).to(
                            self.device
                        ),
                        return_dict=True,
                        output_attentions=False,
                        output_hidden_states=False,
                    )
                except AssertionError:
                    for i, new_token_id in enumerate(new_token_ids):
                        input_ids = torch.tensor([new_token_id]).unsqueeze(0).to(self.device)

                        model_out = self.model_obj(
                            input_ids=input_ids,
                            past_key_values=self._past_key_values,
                            use_cache=True,
                            position_ids=torch.arange(past_length, past_length + 1)
                            .unsqueeze(0)
                            .to(self.device),
                            attention_mask=torch.ones(1, past_length + 1).to(self.device),
                            return_dict=True,
                            output_attentions=False,
                            output_hidden_states=False,
                        )

                        self._past_key_values = model_out.past_key_values
                        past_length += 1

            # save the results
            self._past_key_values = model_out.past_key_values
            cache_token_ids.extend(new_token_ids)
            # Need to add special truncating logic here for weird models that have a different output size than tokenizer vocab
            self._cached_logits = (
                model_out.logits[0, -1, : len(self.tokenizer.tokens)].cpu().numpy()
            )
            self.metrics.engine_input_tokens += len(new_token_ids)
            self.metrics.engine_output_tokens += 1

        return self._cached_logits


class Transformers(Model):
    def __init__(
        self,
        model=None,
        tokenizer=None,
        echo=True,
        compute_log_probs=False,
        chat_template=None,
        **kwargs,
    ):
        """Build a new Transformers model object that represents a model in a given state."""
        super().__init__(
            TransformersEngine(
                model,
                tokenizer,
                compute_log_probs,
                chat_template=chat_template,
                **kwargs,
            ),
            echo=echo,
        )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Grammar(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODES_FIELD_NUMBER: builtins.int
    @property
    def nodes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___GrammarFunction]: ...
    def __init__(
        self,
        *,
        nodes: collections.abc.Iterable[global___GrammarFunction] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["nodes", b"nodes"]) -> None: ...

global___Grammar = Grammar

@typing.final
class EngineCallResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class CaptureGroupsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___Value: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___Value | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing.final
    class CaptureGroupLogProbsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___Value: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___Value | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    NEW_BYTES_FIELD_NUMBER: builtins.int
    IS_GENERATED_FIELD_NUMBER: builtins.int
    NEW_BYTES_PROB_FIELD_NUMBER: builtins.int
    CAPTURE_GROUPS_FIELD_NUMBER: builtins.int
    CAPTURE_GROUP_LOG_PROBS_FIELD_NUMBER: builtins.int
    NEW_TOKEN_COUNT_FIELD_NUMBER: builtins.int
    new_bytes: builtins.bytes
    is_generated: builtins.bool
    new_bytes_prob: builtins.float
    new_token_count: builtins.int
    @property
    def capture_groups(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Value]: ...
    @property
    def capture_group_log_probs(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Value]: ...
    def __init__(
        self,
        *,
        new_bytes: builtins.bytes = ...,
        is_generated: builtins.bool = ...,
        new_bytes_prob: builtins.float = ...,
        capture_groups: collections.abc.Mapping[builtins.str, global___Value] | None = ...,
        capture_group_log_probs: collections.abc.Mapping[builtins.str, global___Value] | None = ...,
        new_token_count: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["capture_group_log_probs", b"capture_group_log_probs", "capture_groups", b"capture_groups", "is_generated", b"is_generated", "new_bytes", b"new_bytes", "new_bytes_prob", b"new_bytes_prob", "new_token_count", b"new_token_count"]) -> None: ...

global___EngineCallResponse = EngineCallResponse

@typing.final
class Value(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STRING_VALUE_FIELD_NUMBER: builtins.int
    BYTES_VALUE_FIELD_NUMBER: builtins.int
    FLOAT_VALUE_FIELD_NUMBER: builtins.int
    LIST_VALUE_FIELD_NUMBER: builtins.int
    string_value: builtins.str
    bytes_value: builtins.bytes
    float_value: builtins.float
    @property
    def list_value(self) -> global___ListValue: ...
    def __init__(
        self,
        *,
        string_value: builtins.str = ...,
        bytes_value: builtins.bytes = ...,
        float_value: builtins.float = ...,
        list_value: global___ListValue | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bytes_value", b"bytes_value", "float_value", b"float_value", "kind", b"kind", "list_value", b"list_value", "string_value", b"string_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bytes_value", b"bytes_value", "float_value", b"float_value", "kind", b"kind", "list_value", b"list_value", "string_value", b"string_value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["kind", b"kind"]) -> typing.Literal["string_value", "bytes_value", "float_value", "list_value"] | None: ...

global___Value = Value

@typing.final
class ListValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[global___Value] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["values", b"values"]) -> None: ...

global___ListValue = ListValue

@typing.final
class Byte(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BYTE_FIELD_NUMBER: builtins.int
    HIDDEN_FIELD_NUMBER: builtins.int
    COMMIT_POINT_FIELD_NUMBER: builtins.int
    NULLABLE_FIELD_NUMBER: builtins.int
    CAPTURE_NAME_FIELD_NUMBER: builtins.int
    TEMPERATURE_FIELD_NUMBER: builtins.int
    byte: builtins.bytes
    hidden: builtins.bool
    commit_point: builtins.bool
    nullable: builtins.bool
    capture_name: builtins.str
    temperature: builtins.float
    def __init__(
        self,
        *,
        byte: builtins.bytes = ...,
        hidden: builtins.bool = ...,
        commit_point: builtins.bool = ...,
        nullable: builtins.bool = ...,
        capture_name: builtins.str = ...,
        temperature: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["byte", b"byte", "capture_name", b"capture_name", "commit_point", b"commit_point", "hidden", b"hidden", "nullable", b"nullable", "temperature", b"temperature"]) -> None: ...

global___Byte = Byte

@typing.final
class ByteRange(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BYTE_RANGE_FIELD_NUMBER: builtins.int
    HIDDEN_FIELD_NUMBER: builtins.int
    COMMIT_POINT_FIELD_NUMBER: builtins.int
    CAPTURE_NAME_FIELD_NUMBER: builtins.int
    TEMPERATURE_FIELD_NUMBER: builtins.int
    byte_range: builtins.bytes
    hidden: builtins.bool
    commit_point: builtins.bool
    capture_name: builtins.str
    temperature: builtins.float
    def __init__(
        self,
        *,
        byte_range: builtins.bytes = ...,
        hidden: builtins.bool = ...,
        commit_point: builtins.bool = ...,
        capture_name: builtins.str = ...,
        temperature: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["byte_range", b"byte_range", "capture_name", b"capture_name", "commit_point", b"commit_point", "hidden", b"hidden", "temperature", b"temperature"]) -> None: ...

global___ByteRange = ByteRange

@typing.final
class Null(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Null = Null

@typing.final
class ModelVariable(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    HIDDEN_FIELD_NUMBER: builtins.int
    COMMIT_POINT_FIELD_NUMBER: builtins.int
    CAPTURE_NAME_FIELD_NUMBER: builtins.int
    NULLABLE_FIELD_NUMBER: builtins.int
    name: builtins.str
    hidden: builtins.bool
    commit_point: builtins.bool
    capture_name: builtins.str
    nullable: builtins.bool
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        hidden: builtins.bool = ...,
        commit_point: builtins.bool = ...,
        capture_name: builtins.str = ...,
        nullable: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["capture_name", b"capture_name", "commit_point", b"commit_point", "hidden", b"hidden", "name", b"name", "nullable", b"nullable"]) -> None: ...

global___ModelVariable = ModelVariable

@typing.final
class Join(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NULLABLE_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    HIDDEN_FIELD_NUMBER: builtins.int
    COMMIT_POINT_FIELD_NUMBER: builtins.int
    CAPTURE_NAME_FIELD_NUMBER: builtins.int
    MAX_TOKENS_FIELD_NUMBER: builtins.int
    nullable: builtins.bool
    name: builtins.str
    hidden: builtins.bool
    commit_point: builtins.bool
    capture_name: builtins.str
    max_tokens: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Use a repeated field to store the list of values"""

    def __init__(
        self,
        *,
        nullable: builtins.bool = ...,
        values: collections.abc.Iterable[builtins.int] | None = ...,
        name: builtins.str = ...,
        hidden: builtins.bool = ...,
        commit_point: builtins.bool = ...,
        capture_name: builtins.str = ...,
        max_tokens: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["capture_name", b"capture_name", "commit_point", b"commit_point", "hidden", b"hidden", "max_tokens", b"max_tokens", "name", b"name", "nullable", b"nullable", "values", b"values"]) -> None: ...

global___Join = Join

@typing.final
class Select(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NULLABLE_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    HIDDEN_FIELD_NUMBER: builtins.int
    COMMIT_POINT_FIELD_NUMBER: builtins.int
    CAPTURE_NAME_FIELD_NUMBER: builtins.int
    MAX_TOKENS_FIELD_NUMBER: builtins.int
    RECURSIVE_FIELD_NUMBER: builtins.int
    nullable: builtins.bool
    name: builtins.str
    hidden: builtins.bool
    commit_point: builtins.bool
    capture_name: builtins.str
    max_tokens: builtins.int
    recursive: builtins.bool
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Use a repeated field to store the list of values"""

    def __init__(
        self,
        *,
        nullable: builtins.bool = ...,
        values: collections.abc.Iterable[builtins.int] | None = ...,
        name: builtins.str = ...,
        hidden: builtins.bool = ...,
        commit_point: builtins.bool = ...,
        capture_name: builtins.str = ...,
        max_tokens: builtins.int = ...,
        recursive: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["capture_name", b"capture_name", "commit_point", b"commit_point", "hidden", b"hidden", "max_tokens", b"max_tokens", "name", b"name", "nullable", b"nullable", "recursive", b"recursive", "values", b"values"]) -> None: ...

global___Select = Select

@typing.final
class GrammarFunction(google.protobuf.message.Message):
    """message Terminal {
        oneof function_type {
            Byte byte = 1;
            ByteRange byte_range = 2;
        }
    }
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOIN_FIELD_NUMBER: builtins.int
    SELECT_FIELD_NUMBER: builtins.int
    BYTE_FIELD_NUMBER: builtins.int
    BYTE_RANGE_FIELD_NUMBER: builtins.int
    MODEL_VARIABLE_FIELD_NUMBER: builtins.int
    @property
    def join(self) -> global___Join: ...
    @property
    def select(self) -> global___Select: ...
    @property
    def byte(self) -> global___Byte: ...
    @property
    def byte_range(self) -> global___ByteRange: ...
    @property
    def model_variable(self) -> global___ModelVariable: ...
    def __init__(
        self,
        *,
        join: global___Join | None = ...,
        select: global___Select | None = ...,
        byte: global___Byte | None = ...,
        byte_range: global___ByteRange | None = ...,
        model_variable: global___ModelVariable | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["byte", b"byte", "byte_range", b"byte_range", "function_type", b"function_type", "join", b"join", "model_variable", b"model_variable", "select", b"select"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["byte", b"byte", "byte_range", b"byte_range", "function_type", b"function_type", "join", b"join", "model_variable", b"model_variable", "select", b"select"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["function_type", b"function_type"]) -> typing.Literal["join", "select", "byte", "byte_range", "model_variable"] | None: ...

global___GrammarFunction = GrammarFunction



import hashlib
import pathlib
import urllib.parse

import diskcache as dc
import platformdirs
import requests

from ._model import Chat
from ._grammarless import GrammarlessEngine, Grammarless


try:
    import openai

    is_openai = True
except ModuleNotFoundError:
    is_openai = False


class AzureAIStudioChatEngine(GrammarlessEngine):
    def __init__(
        self,
        *,
        tokenizer,
        max_streaming_tokens: int,
        timeout: float,
        compute_log_probs: bool,
        azureai_studio_endpoint: str,
        azureai_model_deployment: str,
        azureai_studio_key: str,
        clear_cache: bool,
    ):
        endpoint_parts = urllib.parse.urlparse(azureai_studio_endpoint)
        if endpoint_parts.path == "/score":
            self._is_openai_compatible = False
            self._endpoint = azureai_studio_endpoint
        else:
            if not is_openai:
                raise ValueError(
                    "Detected OpenAI compatible model; please install openai package"
                )
            self._is_openai_compatible = True
            self._endpoint = f"{endpoint_parts.scheme}://{endpoint_parts.hostname}"
        self._deployment = azureai_model_deployment
        self._api_key = azureai_studio_key

        # There is a cache... better make sure it's specific
        # to the endpoint and deployment
        deployment_id = self._hash_prompt(self._endpoint + self._deployment)

        path = (
            pathlib.Path(platformdirs.user_cache_dir("guidance"))
            / f"azureaistudio.tokens.{deployment_id}"
        )
        self.cache = dc.Cache(path)
        if clear_cache:
            self.cache.clear()

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)

    def _hash_prompt(self, prompt):
        # Copied from OpenAIChatEngine
        return hashlib.sha256(f"{prompt}".encode()).hexdigest()

    def _generator(self, prompt: bytes, temperature: float):
        # Initial parts of this straight up copied from OpenAIChatEngine

        # The next loop (or one like it) appears in several places,
        # and quite possibly belongs in a library function or superclass
        # That said, I'm not _completely sure that there aren't subtle
        # differences between the various versions

        assert isinstance(prompt, bytes)

        # find the role tags
        pos = 0
        input_token_count = 0
        role_end = b"<|im_end|>"
        messages = []
        found = True
        while found:

            # find the role text blocks
            found = False
            for role_name, start_bytes in (
                ("system", b"<|im_start|>system\n"),
                ("user", b"<|im_start|>user\n"),
                ("assistant", b"<|im_start|>assistant\n"),
            ):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert (
                            role_name == "assistant"
                        ), "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos : pos + end_pos]
                    pos += end_pos + len(role_end)
                    message_content = btext.decode("utf8")
                    input_token_count += len(self.tokenizer.encode(btext))
                    messages.append({"role": role_name, "content": message_content})
                    found = True
                    break

        # Add nice exception if no role tags were used in the prompt.
        # TODO: Move this somewhere more general for all chat models?
        if messages == []:
            raise ValueError(
                f"The model is a Chat-based model and requires role tags in the prompt! \
            Make sure you are using guidance context managers like `with system():`, `with user():` and `with assistant():` \
            to appropriately format your guidance program for this type of model."
            )

        # Update shared data state
        self._reset_shared_data(prompt[:pos], temperature)

        # Use cache only when temperature is 0
        if temperature == 0:
            cache_key = self._hash_prompt(prompt)

            # Check if the result is already in the cache
            if cache_key in self.cache:
                for chunk in self.cache[cache_key]:
                    yield chunk
                return

        # Call the actual API and extract the next chunk
        if self._is_openai_compatible:
            client = openai.OpenAI(api_key=self._api_key, base_url=self._endpoint)
            response = client.chat.completions.create(
                model=self._deployment,
                messages=messages,  # type: ignore[arg-type]
                # max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                # stream=True,
            )

            result = response.choices[0]
            chunk = result.message.content
            encoded_chunk = chunk.encode("utf8")  # type: ignore[union-attr]

            # Non-streaming OpenAI call, so we can just get the metrics directly
            if response.usage is not None:
                self.metrics.engine_input_tokens += response.usage.prompt_tokens
                self.metrics.engine_output_tokens += response.usage.completion_tokens
        else:
            parameters = dict(temperature=temperature)
            payload = dict(
                input_data=dict(input_string=messages, parameters=parameters)
            )

            headers = {
                "Content-Type": "application/json",
                "Authorization": ("Bearer " + self._api_key),
                "azureml-model-deployment": self._deployment,
            }
            response_score = requests.post(
                self._endpoint,
                json=payload,
                headers=headers,
            )

            result_score = response_score.json()

            chunk = result_score["output"]
            encoded_chunk = chunk.encode("utf8")
            self.metrics.engine_input_tokens += input_token_count
            self.metrics.engine_output_tokens += len(
                self.tokenizer.encode(encoded_chunk)
            )

        # Now back to OpenAIChatEngine, with slight modifications since
        # this isn't a streaming API
        if temperature == 0:
            cached_results = []

        yield encoded_chunk

        if temperature == 0:
            cached_results.append(encoded_chunk)

        # Cache the results after the generator is exhausted
        if temperature == 0:
            self.cache[cache_key] = cached_results


class AzureAIStudioChat(Grammarless, Chat):
    def __init__(
        self,
        azureai_studio_endpoint: str,
        azureai_studio_deployment: str,
        azureai_studio_key: str,
        tokenizer=None,
        echo: bool = True,
        max_streaming_tokens: int = 1000,
        timeout: float = 0.5,
        compute_log_probs: bool = False,
        clear_cache: bool = False,
    ):
        """Create a model object for interacting with Azure AI Studio chat endpoints.

        The required information about the deployed endpoint can
        be obtained from Azure AI Studio.

        A `diskcache`-based caching system is used to speed up
        repeated calls when the temperature is specified to be
        zero.

        Parameters
        ----------
        azureai_studio_endpoint : str
            The HTTPS endpoint deployed by Azure AI Studio
        azureai_studio_deployment : str
            The specific model deployed to the endpoint
        azureai_studio_key : str
            The key required for access to the API
        clear_cache : bool
            Whether to empty the internal cache
        """
        super().__init__(
            AzureAIStudioChatEngine(
                azureai_studio_endpoint=azureai_studio_endpoint,
                azureai_model_deployment=azureai_studio_deployment,
                azureai_studio_key=azureai_studio_key,
                tokenizer=tokenizer,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                clear_cache=clear_cache,
            ),
            echo=echo,
        )



"""Elementary benchmarking for `guidance` development purposes.

`guidance` lives in a fast paced LLM environment, has complex dependencies and is tricky to implement.
These benchmarks are designed to focus on key use cases, where regressions can create havoc.

General guidelines:
- Simplicity first, then customization - reproducibility by the community is encouraged
- Everything takes forever - allow a pathway to scale horizontally
- Goalposts shift - some of the code for benchmarking will change frequently and that's okay

Implementation:

The `bench` function is provided for no frills benchmarking that is designated for
automated testing.

For customization, we provide a notebook demonstration of how to run custom benchmarks
that are near mirror versions of what is available in the `bench` function provided.

Not implemented yet, but we intend to provide an avenue of running the benchmarks via
docker containers that have GPU resourcing to scale horizontally.
"""

from guidance.bench._powerlift import (
    retrieve_langchain,
    langchain_chat_extract_runner,
    langchain_chat_extract_filter_template,
)
from guidance.bench._api import bench

# TODO(nopdive): Enable docker containers to execute benchmarking easily



import typing

import tiktoken

from ._grammarless import Grammarless, GrammarlessEngine

try:
    import openai

    client_class: typing.Optional[typing.Type[openai.OpenAI]] = openai.OpenAI
except ImportError:
    client_class = None


class OpenAIEngine(GrammarlessEngine):
    def __init__(
        self,
        tokenizer,
        max_streaming_tokens,
        timeout,
        compute_log_probs,
        model,
        client_class=client_class,
        **kwargs,
    ):

        if client_class is None:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            )

        self.client = client_class(**kwargs)
        self.model_name = model

        # Create a simple registry of models that use completion endpoints.
        self._completion_models = set(
            [
                "gpt-35-turbo-instruct",
                "gpt-3.5-turbo-instruct",
                "babbage-002",
                "davinci-002",
            ]
        )

        if tokenizer is None:
            tokenizer = tiktoken.encoding_for_model(model)

        super().__init__(tokenizer, max_streaming_tokens, timeout, compute_log_probs)

    def _generator_completion(self, prompt: bytes, temperature: float):
        # Only runs on legacy openAI models that use old completion endpoints.
        self._reset_shared_data(prompt, temperature)  # update our shared data state

        try:
            prompt_decoded = prompt.decode("utf8")
            generator = self.client.completions.create(
                model=self.model_name,
                prompt=prompt_decoded,
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                stream=True,
            )
            self.metrics.engine_input_tokens += len(self.tokenizer.encode(prompt))
        except Exception as e:
            # TODO: add retry logic, but keep token counts straight
            raise e

        for part in generator:
            if len(part.choices) > 0:
                chunk = part.choices[0].text or ""
            else:
                chunk = ""
            self.metrics.engine_output_tokens += len(
                self.tokenizer.encode(chunk.encode())
            )
            yield chunk.encode("utf8")

    def _generator_chat(self, prompt: bytes, temperature: float):
        # find the role tags
        pos = 0
        role_end = b"<|im_end|>\n"
        messages = []
        found = True
        input_token_count = 0

        # TODO: refactor this to method on parent class? (or a util function)
        while found:
            # find the role text blocks
            found = False
            for role_name, start_bytes in (
                ("system", b"<|im_start|>system\n"),
                ("user", b"<|im_start|>user\n"),
                ("assistant", b"<|im_start|>assistant\n"),
            ):
                if prompt[pos:].startswith(start_bytes):
                    pos += len(start_bytes)
                    end_pos = prompt[pos:].find(role_end)
                    if end_pos < 0:
                        assert (
                            role_name == "assistant"
                        ), "Bad chat format! Last role before gen needs to be assistant!"
                        break
                    btext = prompt[pos : pos + end_pos]
                    pos += end_pos + len(role_end)
                    message_content: str = btext.decode("utf8")
                    input_token_count += len(self.tokenizer.encode(btext))
                    messages.append({"role": role_name, "content": message_content})
                    found = True
                    break

        # Add nice exception if no role tags were used in the prompt.
        # TODO: Move this somewhere more general for all chat models?
        if messages == []:
            raise ValueError(
                f"The OpenAI model {self.model_name} is a Chat-based model and requires role tags in the prompt! \
            Make sure you are using guidance context managers like `with system():`, `with user():` and `with assistant():` \
            to appropriately format your guidance program for this type of model."
            )

        # Update shared data state
        self._reset_shared_data(prompt[:pos], temperature)

        # API call and response handling
        try:
            # Ideally, for the metrics we would use those returned by the
            # OpenAI API. Unfortunately, it appears that AzureAI hosted
            # models do not support returning metrics when streaming yet
            generator = self.client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                max_tokens=self.max_streaming_tokens,
                n=1,
                top_p=1.0,  # TODO: this should be controllable like temp (from the grammar)
                temperature=temperature,
                stream=True,
            )
            self.metrics.engine_input_tokens += input_token_count

            for part in generator:
                if len(part.choices) > 0:
                    chunk = part.choices[0].delta.content or ""
                else:
                    chunk = ""
                encoded_chunk = chunk.encode("utf8")
                self.metrics.engine_output_tokens += len(
                    self.tokenizer.encode(encoded_chunk)
                )
                yield encoded_chunk

        except Exception as e:
            # TODO: add retry logic, keeping mind of token counts
            raise e

    def _generator(self, prompt: bytes, temperature: float):
        assert isinstance(prompt, bytes)
        if self.model_name in self._completion_models:
            return self._generator_completion(prompt, temperature)
        else:
            # Otherwise we are in a chat context
            return self._generator_chat(prompt, temperature)


class OpenAI(Grammarless):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        max_streaming_tokens=1000,
        timeout=0.5,
        compute_log_probs=False,
        **kwargs,
    ):
        """Build a new OpenAI model object that represents a model in a given state.

        Parameters
        ----------
        model : str
            The name of the OpenAI model to use (e.g. gpt-3.5-turbo).
        tokenizer : None or tiktoken.Encoding
            The tokenizer to use for the given model. If set to None we use `tiktoken.encoding_for_model(model)`.
        echo : bool
            If true the final result of creating this model state will be displayed (as HTML in a notebook).
        api_key : None or str
            The OpenAI API key to use for remote requests, passed directly to the `openai.OpenAI` constructor.
        max_streaming_tokens : int
            The maximum number of tokens we allow this model to generate in a single stream. Normally this is set very
            high and we rely either on early stopping on the remote side, or on the grammar terminating causing the
            stream loop to break on the local side. This number needs to be longer than the longest stream you want
            to generate.
        **kwargs :
            All extra keyword arguments are passed directly to the `openai.OpenAI` constructor. Commonly used argument
            names include `base_url` and `organization`
        """

        if client_class is None:
            raise Exception(
                "Please install the openai package version >= 1 using `pip install openai -U` in order to use guidance.models.OpenAI!"
            )

        super().__init__(
            engine=OpenAIEngine(
                tokenizer=tokenizer,
                api_key=api_key,
                max_streaming_tokens=max_streaming_tokens,
                timeout=timeout,
                compute_log_probs=compute_log_probs,
                model=model,
                **kwargs,
            ),
            echo=echo,
        )



from json import dumps as json_dumps
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Union,
    Type,
    TYPE_CHECKING,
)

try:
    import jsonschema
    import pydantic
except ImportError:
    if TYPE_CHECKING:
        raise

from .._guidance import guidance
from ..library import char_range, gen, one_or_more, optional, sequence

from .._grammar import GrammarFunction, select, capture, with_temperature
from ._pydantic import pydantic_to_json_schema


def _to_compact_json(target: Any) -> str:
    # See 'Compact Encoding':
    # https://docs.python.org/3/library/json.html
    # Since this is ultimately about the generated
    # output, we don't need to worry about pretty printing
    # and whitespace
    return json_dumps(target, separators=(",", ":"))


class Keyword(str, Enum):
    ANYOF = "anyOf"
    ALLOF = "allOf"
    REF = "$ref"
    CONST = "const"
    ENUM = "enum"
    TYPE = "type"
    PATTERN = "pattern"
    MIN_LENGTH = "minLength"
    MAX_LENGTH = "maxLength"


KEYS = {member.value for member in Keyword}

DEFS_KEYS = {"$defs", "definitions"}

IGNORED_KEYS = {
    "$schema",
    "$id",
    "$comment",
    "title",
    "description",
    "default",
    "examples",
    "required",  # TODO: implement and remove from ignored list
}

TYPE_SPECIFIC_KEYS = {
    "array": {"items", "prefixItems", "minItems", "maxItems"},
    "object": {"properties", "additionalProperties"},
}

STRING_CHARS = [
    char_range("a", "z"),
    char_range("A", "Z"),
    char_range("0", "9"),
    *[c for c in "-_' ,.!?/[]{}():;"],
    "\\n",
    "\\t",
    "\\\\",
]


def validate_json_node_keys(node: Mapping[str, Any]):
    keys = set(node.keys())
    valid_keys = KEYS | IGNORED_KEYS | DEFS_KEYS
    if Keyword.TYPE in node:
        valid_keys |= TYPE_SPECIFIC_KEYS.get(node[Keyword.TYPE], set())
    invalid_keys = keys - valid_keys
    if invalid_keys:
        raise ValueError(
            f"JSON schema had keys that could not be processed: {invalid_keys}" f"\nSchema: {node}"
        )


@guidance(stateless=True)
def _gen_json_int(lm):
    pos_nonzero = char_range("1", "9") + sequence(char_range("0", "9"))
    return lm + optional("-") + select(["0", pos_nonzero])


@guidance(stateless=True)
def _gen_json_number(lm):
    mantissa_int = _gen_json_int()
    mantissa_frac = "." + one_or_more(char_range("0", "9"))
    exponent = "e" + select(["", "+", "-"]) + one_or_more(char_range("0", "9"))

    return lm + mantissa_int + optional(mantissa_frac) + optional(exponent)


@guidance(stateless=True)
def _gen_json_string(
    lm,
    min_length: int = 0,
    max_length: Union[int, None] = None,
    regex: Union[str, None] = None,
):
    lm += '"'
    if regex is not None:
        if min_length > 0 or max_length is not None:
            msg = (
                "If a pattern is specified for a JSON "
                "string, minLength and maxLength must be "
                "left unspecified."
            )
            raise ValueError(msg)
        lm += gen(regex=regex)
    else:
        lm += sequence(select(STRING_CHARS), min_length=min_length, max_length=max_length)
    return lm + '"'


@guidance(stateless=True)
def _gen_json_object(
    lm,
    *,
    properties: Mapping[str, Any],
    additional_properties: Union[bool, Mapping[str, Any]],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    if additional_properties is True:
        # True means that anything goes
        additional_properties = {}

    lm += "{"
    if properties:
        lm += _process_properties(properties=properties, definitions=definitions)
    if properties and additional_properties is not False:
        lm += optional(
            ","
            + _process_additional_properties(
                additional_properties=additional_properties, definitions=definitions
            )
        )
    elif additional_properties is not False:
        lm += optional(
            _process_additional_properties(
                additional_properties=additional_properties, definitions=definitions
            )
        )
    lm += "}"
    return lm


@guidance(stateless=True)
def _process_properties(
    lm,
    *,
    properties: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    properties_added = 0
    for name, property_schema in properties.items():
        lm += '"' + name + '"'

        lm += ":"
        lm += _gen_json(
            json_schema=property_schema,
            definitions=definitions,
        )
        properties_added += 1
        if properties_added < len(properties):
            lm += ","
    return lm


@guidance(stateless=True)
def _process_additional_properties(
    lm,
    *,
    additional_properties: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    item = (
        _gen_json_string()
        + ":"
        + _gen_json(json_schema=additional_properties, definitions=definitions)
    )
    return lm + sequence(item + ",") + item


@guidance(stateless=True)
def _gen_json_array(
    lm,
    *,
    prefix_items_schema: Sequence[Mapping[str, Any]],
    item_schema: Union[bool, Mapping[str, Any]],
    min_items: int,
    max_items: Optional[int],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    if item_schema is True:
        # True means that anything goes
        item_schema = {}

    if len(prefix_items_schema) < min_items and item_schema is False:
        raise ValueError(
            f"PrefixItems has too few elements ({len(prefix_items_schema)}) to"
            f" satisfy minItems ({min_items}) but no extra items were allowed"
        )

    if max_items is not None and max_items < min_items:
        raise ValueError(f"maxItems ({max_items}) can't be less than minItems ({min_items})")

    required_items = []
    optional_items = []

    # If max_items is None, we can add an infinite tail of items later
    n_to_add = max(len(prefix_items_schema), min_items) if max_items is None else max_items
    for i in range(n_to_add):
        if i < len(prefix_items_schema):
            schema = prefix_items_schema[i]
        elif item_schema is not False:
            schema = item_schema
        else:
            assert i >= min_items
            break

        item = _gen_json(json_schema=schema, definitions=definitions)

        if i < min_items:
            required_items.append(item)
        else:
            optional_items.append(item)

    if max_items is None and item_schema is not False:
        # Add an infinite tail of items
        item = _gen_json(json_schema=item_schema, definitions=definitions)
        optional_items.append(item + sequence("," + item))

    lm += "["

    if required_items:
        first, *rest = required_items
        lm += first
        for item in rest:
            lm += "," + item

    if optional_items:
        # This is a bit subtle and would not be required if not for prefixItems -- the previous
        # must be present before the next one may be added, meaning we have nested optionals:
        # (first optional(,second optional(,third (optional(,...)))))
        first, *rest = optional_items
        tail = ""
        for item in reversed(rest):
            tail = optional("," + item + tail)
        tail = first + tail

        if required_items:
            lm += optional("," + tail)
        else:
            lm += optional(tail)

    lm += "]"
    return lm


@guidance(stateless=True)
def _process_anyOf(
    lm,
    *,
    anyof_list: Sequence[Mapping[str, Any]],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    options = [_gen_json(json_schema=item, definitions=definitions) for item in anyof_list]
    return lm + select(options)


@guidance(stateless=True)
def _process_enum(lm, *, options: Sequence[Mapping[str, Any]]):
    # options will come in as python objects, so we need to convert to (compact) JSON
    all_opts = []
    for opt in options:
        all_opts.append(_to_compact_json(opt))
    return lm + select(options=all_opts)


@guidance(stateless=True)
def _gen_json_any(lm):
    return lm + select(
        [
            _gen_json(json_schema={"type": "null"}, definitions={}),
            _gen_json(json_schema={"type": "boolean"}, definitions={}),
            _gen_json(json_schema={"type": "integer"}, definitions={}),
            _gen_json(json_schema={"type": "number"}, definitions={}),
            _gen_json(json_schema={"type": "string"}, definitions={}),
            # Recursive cases
            _gen_json(
                json_schema={
                    "type": "array",
                    "items": True,
                },
                definitions={},
            ),
            _gen_json(
                json_schema={
                    "type": "object",
                    "additionalProperties": True,
                },
                definitions={},
            ),
        ]
    )


@guidance(stateless=True)
def _gen_json(
    lm,
    json_schema: Mapping[str, Any],
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    validate_json_node_keys(json_schema)

    if Keyword.ANYOF in json_schema:
        return lm + _process_anyOf(anyof_list=json_schema[Keyword.ANYOF], definitions=definitions)

    if Keyword.ALLOF in json_schema:
        allof_list = json_schema[Keyword.ALLOF]
        if len(allof_list) != 1:
            raise ValueError("Only support allOf with exactly one item")
        return lm + _gen_json(allof_list[0], definitions)

    if Keyword.REF in json_schema:
        return lm + _get_definition(reference=json_schema[Keyword.REF], definitions=definitions)

    if Keyword.CONST in json_schema:
        return lm + _to_compact_json(json_schema[Keyword.CONST])

    if Keyword.ENUM in json_schema:
        return lm + _process_enum(options=json_schema[Keyword.ENUM])

    if Keyword.TYPE in json_schema:
        target_type = json_schema[Keyword.TYPE]
        if target_type == "null":
            return lm + "null"
        if target_type == "boolean":
            return lm + select(["true", "false"])
        if target_type == "integer":
            return lm + _gen_json_int()
        if target_type == "number":
            return lm + _gen_json_number()
        if target_type == "string":
            return lm + _gen_json_string(
                regex=json_schema.get(Keyword.PATTERN, None),
                min_length=json_schema.get(Keyword.MIN_LENGTH, 0),
                max_length=json_schema.get(Keyword.MAX_LENGTH, None),
            )
        if target_type == "array":
            return lm + _gen_json_array(
                prefix_items_schema=json_schema.get("prefixItems", []),
                item_schema=json_schema.get("items", True),
                min_items=json_schema.get("minItems", 0),
                max_items=json_schema.get("maxItems"),
                definitions=definitions,
            )
        if target_type == "object":
            return lm + _gen_json_object(
                properties=json_schema.get("properties", {}),
                additional_properties=json_schema.get("additionalProperties", True),
                definitions=definitions,
            )
        raise ValueError(f"Unsupported type in schema: {target_type}")

    return lm + _gen_json_any()


@guidance(stateless=True)
def json(
    lm,
    name: Optional[str] = None,
    *,
    schema: Union[
        None,
        Mapping[str, Any],
        Type["pydantic.BaseModel"],
        "pydantic.TypeAdapter",
    ] = None,
    temperature: float = 0.0,
):
    """Generate valid JSON according to the supplied JSON schema or `pydantic` model.

    Not all parts of `JSON schema <https://json-schema.org/>`_ are supported. Indeed some parts
    (such as bounds on numbers) cannot really be supported in the context of LLM generation.

    Using a JSON schema:

        >>> schema = ''{ "type": "object", "properties": { "a" : {"type": "integer"} } }'
        >>> schema_obj = json.loads(schema)
        >>> lm += json(name="generated_object", schema=schema_obj)
        >>> print(json.loads(lm["generated_object"]))
        { 'a' : 2 }

    Using a ``pydantic.BaseModel``:

        >>> class Schema(BaseModel):
        ...     b: bool
        >>> lm += json(name="generated_object", schema=Schema)
        >>> print(json.loads(lm["generated_object"]))
        { 'b' : False }

    Using a ``pydantic.TypeAdapter``:

        >>> schema = TypeAdapter(list[int])
        >>> lm += json(name="generated_object", schema=schema)
        >>> print(json.loads(lm["generated_object"]))
        [1, 2, 3]

    Parameters
    ----------

    name : str or None
        If this is not None then the the results of the generation will be saved as a variable on
        the Model object (so you can access the result as ``lm["var_name"]``).

    schema : Union[None, Mapping[str, Any], Type[pydantic.BaseModel], pydantic.TypeAdapter]
        One of:
            - None, in which case any valid JSON will be generated
            - A JSON schema object. This is a JSON schema string which has been passed to ``json.loads()``
            - A subclass of ``pydantic.BaseModel``
            - An instance of ``pydantic.TypeAdapter``
    """
    if isinstance(schema, Mapping):
        # Raises jsonschema.exceptions.SchemaError or ValueError
        # if schema is not valid
        jsonschema.validators.Draft202012Validator.check_schema(schema)
    elif schema is None:
        schema = {}
    else:
        schema = pydantic_to_json_schema(schema)

    definitions: Mapping[str, Callable[[], GrammarFunction]] = {}
    for dk in DEFS_KEYS:
        if dk in schema:
            assert len(definitions) == 0, "Found duplicate definitions"
            definitions = _build_definitions(schema[dk])

    return lm + capture(
        with_temperature(_gen_json(schema, definitions), temperature=temperature),
        name=name,
    )


def _build_definitions(
    raw_definitions: Mapping[str, Any]
) -> Mapping[str, Callable[[], GrammarFunction]]:
    definitions: Dict[str, Callable[[], GrammarFunction]] = {}

    def build_definition(json_schema: Mapping[str, Any]) -> Callable[[], GrammarFunction]:
        @guidance(stateless=True, dedent=False)
        def closure(lm):
            return lm + _gen_json(json_schema=json_schema, definitions=definitions)

        return closure

    definitions = {ref: build_definition(schema) for ref, schema in raw_definitions.items()}
    return definitions


@guidance(stateless=True)
def _get_definition(
    lm,
    *,
    reference: str,
    definitions: Mapping[str, Callable[[], GrammarFunction]],
):
    assert definitions is not None
    target_definition = None
    for dk in DEFS_KEYS:
        ref_start = f"#/{dk}/"
        if reference.startswith(ref_start):
            target_name = reference[len(ref_start) :]
            target_definition = definitions[target_name]

    assert target_definition is not None
    return lm + target_definition()



from typing import Optional

from .._guidance import guidance

# from ._prefix_tree import prefix_tree
from .._grammar import string, select, capture
from ._optional import optional


# @guidance(stateless=True, dedent=False)
# def substring(lm, s):

#     # build a prefix tree over all suffixes
#     strings = [s[i:] for i in range(len(s))]
#     return lm + prefix_tree(strings, partial_matches=True)


@guidance(stateless=True, dedent=False)
def substring_simple(lm, s):
    grammars = ["" for _ in range(len(s))]
    grammars[-1] = select(["", string(s[-1])])
    for i in range(len(s) - 2, -1, -1):
        si = string(s[i])
        grammars[i] = select([si, si + grammars[i + 1]])
    return lm + select(grammars)


class State:
    def __init__(self):
        self.len = 0
        self.link = -1
        self.next = {}


class SuffixAutomaton:
    """A suffix automoton.

    For details see https://en.wikipedia.org/wiki/Suffix_automaton.

    Implementation is based on https://cp-algorithms.com/string/suffix-automaton.html
    """

    def __init__(self, s):

        # init our variables
        self.MAXLEN = 2 * len(
            s
        )  # we can't have more than 2n-1 states for a string of length n
        self.states = [State() for _ in range(self.MAXLEN * 2)]
        self.states[0].len = 0
        self.states[0].link = -1
        self.sz = 1  # init with a single state
        self.last = 0

        # Add characters to the suffix automaton
        for character in s:
            self.sa_extend(character)

    @property
    def root(self):
        return self.states[0]

    def sa_extend(self, c):
        cur = self.sz
        self.sz += 1
        self.states[cur].len = self.states[self.last].len + 1

        p = self.last
        while p != -1 and c not in self.states[p].next:
            self.states[p].next[c] = cur
            p = self.states[p].link

        if p == -1:
            self.states[cur].link = 0
        else:
            q = self.states[p].next[c]
            if self.states[p].len + 1 == self.states[q].len:
                self.states[cur].link = q
            else:
                clone = self.sz
                self.sz += 1
                self.states[clone].len = self.states[p].len + 1
                self.states[clone].next = self.states[q].next.copy()
                self.states[clone].link = self.states[q].link

                while p != -1 and self.states[p].next.get(c) == q:
                    self.states[p].next[c] = clone
                    p = self.states[p].link

                self.states[q].link = self.states[cur].link = clone

        self.last = cur


@guidance(stateless=True, dedent=False)
def substring(lm, target_string: str, name: Optional[str] = None):
    suffix_automaton = SuffixAutomaton(target_string)
    node_cache = {}
    state_stack = [0]  # Start with the initial state index (0) on the stack

    # Loop as long as there are states on the stack
    while state_stack:
        state_ind = state_stack[-1]  # Check the state on the top of the stack

        state = suffix_automaton.states[state_ind]

        # If we have already computed the result for this state, skip it
        if state_ind in node_cache:
            state_stack.pop()
            continue

        # If the state is a leaf node, meaning no outgoing edges (is an end of some suffix)
        if not state.next:
            node_cache[state_ind] = string(
                ""
            )  # Leaf nodes represent empty string suffixes
            state_stack.pop()
            continue

        # If there's an unprocessed child, add it to the stack
        unprocessed_children = [
            next_state
            for next_state in state.next.values()
            if next_state not in node_cache
        ]
        if unprocessed_children:
            state_stack.extend(unprocessed_children)
        else:
            # Once all children are processed, create the node for this state
            options = [string(c) + node_cache[state.next[c]] for c in state.next]
            node_cache[state_ind] = (
                optional(select(options, skip_checks=True))
                if len(options) > 1
                else optional(options[0])
            )
            state_stack.pop()

    return lm + capture(node_cache[0], name=name)


# @guidance(stateless=True, dedent=False)
# def substring(s):
#     a = SuffixAutomaton(s)
#     return _rec_substring(a, 0, {})

# def _rec_substring(suffix_automaton, state_ind, node_cache):
#     if state_ind in node_cache:
#         return node_cache[state_ind]

#     state = suffix_automaton.states[state_ind]

#     if len(state.next) == 0:
#         return string("")

#     options = []
#     for c in state.next:
#         options.append(string(c) + _rec_substring(suffix_automaton, state.next[c], node_cache))

#     if len(options) == 1:
#         node = optional(options[0])
#     else:
#         node = optional(select(options, skip_checks=True))

#     node_cache[state_ind] = node
#     return node



from typing import Sequence

import numpy as np

from ._model import Engine, Model, Chat
from ._remote import RemoteEngine
from ._tokenizer import Tokenizer


class MockTokenizer(Tokenizer):
    def __init__(self, tokens: Sequence[bytes]):

        super().__init__(tokens, chat_template=None, bos_token_id=0, eos_token_id=0)

    def recode(self, tokens: Sequence[int]) -> Sequence[int]:
        # Make a no-op for now
        return tokens


class MockEngine(Engine):
    def __init__(self, tokenizer, byte_patterns, compute_log_probs, force):
        super().__init__(tokenizer, compute_log_probs=compute_log_probs)

        self._valid_mask = np.zeros(len(tokenizer.tokens))
        for i, t in enumerate(tokenizer.tokens):
            try:
                t.decode("utf8")
                self._valid_mask[i] = 1.0
            except:
                pass
        self.force = force
        self.called_temperatures = []

        # allow a single byte pattern to be passed
        if isinstance(byte_patterns, (bytes, str)):
            byte_patterns = [byte_patterns]

        # allow for strings to be passed
        for i, pattern in enumerate(byte_patterns):
            if isinstance(pattern, str):
                byte_patterns[i] = pattern.encode("utf8")

        self.byte_patterns = byte_patterns

        # seed the random number generator
        self._rand_generator = np.random.default_rng(seed=42)

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """Pretends to compute the logits for the given token state."""
        self.called_temperatures.append(current_temp)

        # build the byte strings
        byte_string = b"".join(self.tokenizer.tokens[i] for i in token_ids)

        # if we are forcing the bytes patterns then don't allow other tokens
        if self.force:
            logits = np.ones(len(self.tokenizer.tokens)) * -np.inf

        # otherwise we randomly generate valid unicode bytes
        else:
            logits = (
                self._rand_generator.standard_normal(len(self.tokenizer.tokens))
                * self._valid_mask
            )

        # if we have a pattern that matches then force the next token
        bias = 100.0
        if self.byte_patterns is not None:
            byte_string
            for p in self.byte_patterns:
                if p.startswith(byte_string) and len(p) > len(byte_string):
                    for i in self._get_next_tokens(p[len(byte_string) :]):
                        logits[i] += bias
                    bias /= 2  # if we have multiple matches then they apply with decreasing bias

        return logits

    def _get_next_tokens(self, byte_string):
        for i, t in enumerate(self.tokenizer.tokens):
            if byte_string.startswith(t):
                yield i


class Mock(Model):
    def __init__(
        self,
        byte_patterns=[],
        echo=True,
        compute_log_probs=False,
        force=False,
        **kwargs,
    ):
        """Build a new Mock model object that represents a model in a given state."""

        if isinstance(byte_patterns, str) and byte_patterns.startswith("http"):
            engine = RemoteEngine(byte_patterns, **kwargs)
        else:
            # Our tokens are all bytes and all lowercase letter pairs
            all_lc_pairs = [
                bytes([i, j])
                for i in range(ord("a"), ord("z"))
                for j in range(ord("a"), ord("z"))
            ]
            all_bytes = [bytes([i]) for i in range(256)]
            tokens = [b"<s>"] + all_lc_pairs + all_bytes

            tokenizer = MockTokenizer(tokens)
            engine = MockEngine(tokenizer, byte_patterns, compute_log_probs, force)

        super().__init__(engine, echo=echo)


class MockChat(Mock, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from .._guidance import guidance
from ._any_char import any_char
from .._grammar import select, capture, string, commit_point
from ._sequences import zero_or_more, one_or_more
from ._any_char_but import any_char_but
from ._any_char import any_char


class Tool:
    def __init__(self, call_grammar=None, tool_call=None, callable=None):
        # call_grammar specifies how the tool can be called. Crucially, it has to capture the args in variable 'tool_args'
        # tool_call is a guidance function  actually calls the tool, and returns an lm object with whatever outputs it wants
        # callable: guidance function or regular callable, will be converted to grammar
        # TODO: hidden is not working yet
        first_option = (call_grammar is not None) and (tool_call is not None)
        second_option = callable is not None
        # either both are true or both false
        if first_option == second_option:
            raise Exception(
                "Must pass either (call_grammar, tool call) or callable, but not both or neither"
            )
        if second_option:
            call_grammar, tool_call = fn_to_grammar_call(callable)
        self.call_grammar = call_grammar
        self.tool_call = tool_call


def valid_chars():
    return any_char_but(["=", ")"])


def positional_arg():
    return one_or_more(valid_chars())


def kwarg():
    return one_or_more(valid_chars()) + "=" + one_or_more(valid_chars())


def basic_func_grammar(name):
    obj = string(name + "(")
    obj += capture(
        select([zero_or_more(positional_arg()), ""]) + select([zero_or_more(kwarg()), ""]),
        name="tool_args",
    )
    obj += string(")")
    return obj


def fn_to_grammar_call(callable):
    # TODO later: validate the call. Here is code to get required and optional args of 'guidance_fn':
    # name = guidance_fn.__name__
    # required_args = []
    # optional_args = []
    # sig = inspect.signature(guidance_fn)
    # for i, x in enumerate(sig.parameters.values()):
    #     if i == 0:
    #         continue
    #     if x.default is x.empty:
    #         required_args.append(x.name)
    #     else:
    #         optional_args.append(x.name)
    name = callable.__name__
    call_grammar = basic_func_grammar(name)

    @guidance(dedent=False)
    def basic_tool_call(lm):
        args = lm["tool_args"]
        args = args.split(",")
        positional = [x.strip() for x in args if "=" not in x]
        kwargs = dict([tuple(x.strip().split("=")) for x in args if "=" in x])
        lm += callable(*positional, **kwargs)
        return lm

    return call_grammar, basic_tool_call



from .._guidance import guidance
from .._grammar import Byte, byte_range, select


@guidance(stateless=True)
def any_char_but(lm, forbidden):
    """Allows any char except those in forbidden"""
    # TODO: extend this to support utf-8 encoded multibyte unicode characters
    forb = sorted(set(ord(x) for x in forbidden))
    # Add end value to simplify loop (assumes 0-127 ascii)
    forb.append(128)
    start = 0
    ranges = []
    singletons = []
    for i in forb:
        if start < i - 1:
            # Add range if there is more than one character
            ranges.append((start, i - 1))
        if start == i - 1:
            # Add singleton if there is only one character
            singletons.append(start)
        start = i + 1
    ranges = [byte_range(bytes([i]), bytes([j])) for i, j in ranges]
    singletons = [Byte(bytes([i])) for i in singletons]
    return select(ranges + singletons)



from ._model import Model, Instruct, Chat

# local models
from .transformers._transformers import Transformers, TransformersTokenizer
from .llama_cpp import LlamaCpp
from ._mock import Mock, MockChat

# grammarless models (we can't do constrained decoding for them)
from ._grammarless import Grammarless
from .vertexai._vertexai import (
    VertexAI,
    VertexAIChat,
    VertexAICompletion,
    VertexAIInstruct,
)
from ._azure_openai import (
    AzureOpenAI,
)
from ._azureai_studio import AzureAIStudioChat
from ._openai import OpenAI
from ._lite_llm import LiteLLM, LiteLLMChat, LiteLLMInstruct, LiteLLMCompletion
from ._cohere import Cohere, CohereCompletion, CohereInstruct
from ._anthropic import Anthropic
from ._googleai import GoogleAI, GoogleAIChat
from ._togetherai import (
    TogetherAI,
    TogetherAIChat,
    TogetherAIInstruct,
    TogetherAICompletion,
)

# server based grammar models
from ._azure_guidance import AzureGuidance
